from yann.data.classes import Classes



def test_classes():
  classes = Classes.ordered(10)

  assert len(classes) == 10
  assert 1 in classes
  assert classes[1] == 1

  classes = Classes(['a', 'b', 'c'])

  assert 'a' in classes
  assert classes.indices['a'] == 0
  assert classes.index_encode('a') == 0
  assert classes.one_hot_encode('a')[0] == 1


  classes = Classes(counts={'a': 10, 'b': 20})
  assert classes.weights() == [30/10, 30/20]


import yann

class files:
  class_names = ''


class OpenImages:
  """



  https://storage.googleapis.com/openimages/web/download.html
  """




  def __init__(self, root=None, classes=None, download=True):
    self.images = []


  def image_url(self, id):
    pass

  def image_path(self, id):
    pass

  def download(self):
    for id in self.images:
      url = self.image_url(id)
      yann.download(url)


class OpenImagesExtended(OpenImages):
  """
  https://storage.googleapis.com/openimages/web/extended.html
  """


from torch import nn
import torch
from yann.typedefs import Logits, MultiLabelOneHot

from yann.modules.loss import _reduce


class AsymmetricLoss(nn.Module):
  """
  ASL - Asymmetric Loss for Multi Label Classification
  https://arxiv.org/abs/2009.14119v1

  m = probability margin
  shifted probability:
    p_m = max(p - m, 0)

  ```math
  A S L=\left\{\begin{array}{ll}
  L_{+}= & (1-p)^{\gamma_{+}} \log (p) \\
  L_{-}= & \left(p_{m}\right)^{\gamma_{-}} \log \left(1-p_{m}\right)
  \end{array}\right.
  ```

  based on official implementation:
    https://github.com/Alibaba-MIIL/ASL
  """

  def __init__(self, neg_decay=4, pos_decay=1, prob_shift=0.05, eps=1e-8, reduction='mean'):
    super(AsymmetricLoss, self).__init__()
    self.pos_decay = pos_decay
    self.neg_decay = neg_decay
    self.prob_shift = prob_shift or 0

    self.eps = eps

    self.reduction = reduction
    self.calculate_focal_loss_gradients = True


  def forward(self, inputs: Logits, targets: MultiLabelOneHot):
    pos_probs = torch.sigmoid(inputs)
    neg_probs = 1 - pos_probs

    if self.prob_shift > 0:
      # negative shifting max(p - shift, 0),
      # effectively ignores easy negative examples with small loss
      # paper claims this also helps handling mislabeled negative examples
      neg_probs = (neg_probs + self.prob_shift).clamp(max=1)

    neg_targets = (1 - targets)

    pos_losses = targets * torch.log(pos_probs.clamp(min=self.eps))
    neg_losses = neg_targets * torch.log(neg_probs.clamp(min=self.eps))
    losses = pos_losses + neg_losses  # standard cross entropy

    if self.pos_decay or self.neg_decay:
      with torch.set_grad_enabled(self.calculate_focal_loss_gradients):
        pts = pos_probs * targets + neg_probs * neg_targets
        decays = self.pos_decay * targets + self.neg_decay * neg_targets
        weights = torch.pow(1 - pts, decays)
      losses *= weights

    return _reduce(-losses, reduction=self.reduction)


class AsymmetricLossOptimized(AsymmetricLoss):
  def forward(self, inputs: Logits, targets: MultiLabelOneHot):
    self.targets = targets
    self.neg_targets = (1 - self.targets)

    self.pos_probs = torch.sigmoid(inputs)
    self.neg_probs = 1 - self.pos_probs

    if self.prob_shift > 0:
      self.neg_probs = self.neg_probs.add_(self.prob_shift).clamp_(max=1)

    self.losses = self.targets * torch.log(self.pos_probs)
    self.losses.add_(self.neg_targets * torch.log(self.neg_probs))

    if self.pos_decay or self.neg_decay:
      with torch.set_grad_enabled(self.calculate_focal_loss_gradients):
        self.pos_probs.mul_(self.targets)
        self.neg_probs.mul_(self.neg_targets)
        weights = torch.pow(
          1 - self.pos_probs - self.neg_probs,
          self.pos_decay * self.targets + self.neg_decay * self.neg_targets
        )
      self.losses *= weights

    return _reduce(-self.losses, reduction=self.reduction)


import torch
import typing


Logits = torch.FloatTensor
Embedding = torch.FloatTensor
OneHot = torch.Tensor

Probabilities = torch.FloatTensor
LogProbabilities = torch.FloatTensor

Batch = typing.NewType('Batch', torch.Tensor)

ClassIndices = torch.LongTensor
OneHot = torch.FloatTensor
MultiLabelOneHot = OneHot

ImageTensor = torch.ShortTensor
NormalizedImageTensor = torch.FloatTensor

RGB = ImageTensor
Grayscale = ImageTensor



from .supervised import Trainer


class DistilledTrainer(Trainer):
  def __init__(self, student, teacher, **kwargs):
    super(DistilledTrainer, self).__init__(model=student, **kwargs)
    self.teacher = teacher

  @property
  def student(self):
    return self.model



from torch import nn


class EfficientChannelAttention(nn.Module):
  """
  https://github.com/BangguWu/ECANet
  """
  def __init__(self, kernel_size=3):
    super().__init__()
    self.pool = nn.AdaptiveAvgPool2d(1)
    self.conv = nn.Conv1d(
      1, 1,
      kernel_size=kernel_size,
      padding=(kernel_size - 1) // 2,
      bias=False
    )
    self.sigmoid = nn.Sigmoid()

  def forward(self, input):
    x = self.pool(input)
    x = self.conv(x.squeeze(-1).transpose(-1, -2)).transpose(-1, -2).unsqueeze(-1)
    x = self.sigmoid(x)

    return input * x.expand_as(input)

ECA = EfficientChannelAttention



"""
TODO: way to wrap a function to parametrize it with a random function or argument generator,
 track parameters used and outputs



augment = partial(resize, width=N(0,1))

outputs, params = augment.parametrized(image)
outputs = augment(image)
outputs = augment(image, width=20)


augment = partial(resize, width=Schedule())
augment = partial(resize, width=Compute())  # compute param as a function of provided params

augment.width = 3

"""
# class Partial:
#   def __init__(self, func, **kwargs):
#     self.func = func
#     self.kwargs = kwargs
#
#   def __call__(self, *args, **kwargs):
#     return self.func(*args, **kwargs)
#
# resize = yann.partial(resize, width=N(0 ,1), U(0 ,10))


from dataclasses import dataclass
from typing import Any, Optional

from yann import to


@dataclass
class Batch:
  ids: Optional[Any] = None
  inputs: Optional[Any] = None
  targets: Optional[Any] = None
  outputs: Optional[Any] = None
  losses: Optional[Any] = None

  @property
  def size(self):
    if self.inputs is not None:
      return len(self.inputs)
    raise ValueError('Could not determine size')

  def __iter__(self):
    yield self.inputs
    yield self.outputs

  def to(self, device, **kwargs):
    to(
      self.inputs,
      self.targets,
      self.outputs,
      self.losses,
      device=device,
      **kwargs
    )
    return self


batch = Batch()
batch.size

for batch in batches:
  batch.to('cuda:0')


class Batch:
  inputs: Inputs
  targets: Targets

  def __init__(self, inputs, targets, outputs=None):
    self.inputs = inputs
    self.targets = targets

  def __iter__(self):
    return (*self.inputs, *self.targets)

  def to(self, device):
    pass

  @property
  def size(self):
    return len(self.inputs[0])

  def items(self):
    return []


class DetectionBatch(Batch):
  boxes: Inputs


class SegmentationBatch(Batch):
  masks: Inputs


batches: List[Batch] = []
for b in batches:
  b.to(model.device)
  inputs, targets = b

  outputs = model(*inputs)
  b.outputs = outputs



from typing import Union, Callable, Iterable
import time
from torch import nn
from torch.utils.data.dataset import Dataset
from torch.utils.data.dataloader import DataLoader
import torch.jit
import torch
import yann

from ..data.loaders import TransformLoader


def inference_stream(
    model: Union[nn.Module, Callable, str],
    data: Union[Dataset, DataLoader, Iterable, str],
    device=None,
    transform=None,
    batch_size=64,
    parallel=False,
    num_workers=1,
    pin_memory=False,
    shuffle=False,
    progress=10,
    eval=True,
):
  device = device or yann.default.device

  if isinstance(model, str):
    model = torch.jit.load(model, 'cpu')

  if isinstance(model, nn.Module):
    if parallel:
      model = nn.DataParallel(model)
    if eval: model.eval()
    model.to(device)

  if isinstance(data, str):
    data = yann.resolve.dataset(data)

  if isinstance(data, Dataset):
    data = TransformLoader(
      data,
      transform=transform,
      pin_memory=pin_memory,
      batch_size=batch_size,
      shuffle=shuffle,
      num_workers=num_workers
    )

  try:
    size = len(data)
  except:
    size = -1

  start = begin = time.time()

  with torch.no_grad():
    for idx, (inputs, *rest) in enumerate(data):
      inputs = inputs.to(device)
      outputs = model(inputs)

      yield (inputs, *rest, outputs)

      if progress and idx % progress == 0:
        print(f"[{idx} / {size}] ({time.time() - begin}, total: {time.time() - start})")

      begin = time.time()




"""



TODO: snapshot testing
TODO: acceptance criteria / validation against test set
"""


import torch
from contextlib import contextmanager

from .utils.debug import iter_allocated_tensors
from .exceptions import CheckFailure


def check_tensor(
  t: torch.Tensor,
  device=None,
  dtype=None,
  requires_grad=None,
  contiguous=None,
  pinned=None,
  share_memory=None,
  not_share_memory=None,
  shape=None,
  same=None,
  different=None,
  anomalies=True,
  equal=None,
  close=None,
  like=None,
  lt=None,
  gt=None,
  lte=None,
  gte=None,
  none=False
):
  if share_memory is not None:
    assert t.storage().data_ptr() == share_memory.storage().data_ptr()
  if not_share_memory is not None:
    assert t.storage().data_ptr() != not_share_memory.storage(
    ).data_ptr()
  if different is not None:
    assert different is not t
  if same is not None:
    assert same is t
  if like is not None:
    check_tensor(
      t, device=like.device, shape=like.shape, dtype=like.shape
    )
  if not none:
    assert t is not None
  if device:
    if isinstance(device, str):
      device = torch.device(device)
    assert t.device == device
  if dtype:
    assert t.dtype == dtype
  if requires_grad is not None:
    assert t.requires_grad == requires_grad
  if contiguous is not None:
    assert t.is_contiguous() == contiguous
  if pinned is not None:
    assert t.is_pinned() == pinned
  if shape:
    for x, y in zip(t.shape, shape):
      if y is not None:
        assert x == y
    assert len(t.shape) == len(shape)
  if anomalies:
    assert not torch.isnan(t).any() and not torch.isinf(t).any()
  if equal is not None:
    if not torch.is_tensor(equal):
      equal = torch.Tensor(equal, dtype=t.dtype, device=t.device)
    assert (t == equal).all()
  if close is not None:
    if not torch.is_tensor(close):
      close = torch.Tensor(close, dtype=t.dtype, device=t.device)
    assert torch.allclose(t, close)
  if lt is not None:
    assert (t < lt).all()
  if lte is not None:
    assert (t <= lte).all(), lte
  if gt is not None:
    assert (t > gt).all(), t > gte
  if gte is not None:
    assert (t >= gte).all(), t >= gte


@contextmanager
def newly_allocated_tensors(count=None, max=None):
  """
  Counts the number of allocated tensors at start and exit of block.
  Args:
    count: exact count
    max: at most this many
    new: if True will could newly allocated, otherwise will compare

  Returns:

  """

  initial = set(iter_allocated_tensors())
  new_tensors = set()
  yield new_tensors
  allocated = set(iter_allocated_tensors())
  new_tensors.update(allocated - initial)
  diff = len(new_tensors)

  if count is not None and count != diff:
    raise CheckFailure(
      f'Expected {count} tensor allocations but got {diff}'
    )
  if max is not None:
    raise CheckFailure(
      f'Expected at most {count} tensor allocations but got {diff}'
    )


class Checker:
  def newly_allocated_tensors(self, count=None, max=None):
    return newly_allocated_tensors(count=count, max=max)


# def profile(name=None, sync=True, max=None, track=False, log=True):
#   """
#   track execution speed of code block
#   Args:
#     sync: cuda synchronize or not
#     max: fail if takes longer than x amount of time
#     name:
#     track:
#
#   Returns:
#
#   """
#   pass
#
#
# def track_value():
#   pass

#
# with profile('forward pass', max=.3):
#   pass

# def assert_finite(x):
#   pass
#
#
# def check_stability():
#   pass
#
#
# def check_numerical_issues(function, shapes, types):
#   pass


def rand_image_tensor(
  height,
  width=None,
  channels=3,
  min=0,
  max=1,
  dtype=None,
  device=None
):
  if width is None:
    width = height
  t = torch.rand(channels, height, width, dtype=dtype, device=device)
  return (min - max) * t + max


def rand_image_batch(
  height,
  width=None,
  channels=3,
  num=16,
  min=0,
  max=1,
  dtype=None,
  device=None
):
  if width is None:
    width = height
  t = torch.rand(
    num, channels, height, width, dtype=dtype, device=device
  )
  return (min - max) * t + max


def check_model(
  model,
  input=None,
  output=None,
  input_shape=None,
  output_shape=None,
  # loss=None,
  device=None,
  dtype=None
):
  device = device or model.device
  input = input if input is not None else torch.rand(
    input_shape, device=device, dtype=dtype
  )

  output = model(input)
  check_tensor(output, shape=output_shape, equal=output)



def optimize(
  model=None,
  benchmark=True,
  jit=True,
  amp=True,
  xla=False,
  quantize=False,
  freeze=False,
  example=None,
  device=None
):
  if benchmark:
    from torch.backends import cudnn
    cudnn.benchmark = True

  if model:
    if example:
      if isinstance(example, tuple):
        import torch
        example = torch.randn(*example)

      if jit:
        import torch.jit
        model = torch.jit.trace(model, example)
  return model






from .base import Callback


class ProgressBar(Callback):
  def __init__(self, length=None, samples=True, notebook=False):
    """
    Args:
      length: number of steps per epoch, will infer from dataset or loader if not provided
      samples: count samples, if False will track batches/steps
      notebook: use notebook progress bar
    """
    self.notebook = notebook

    self.length = length
    self.samples = samples

  def on_epoch_start(self, epoch=None, trainer=None):
    if self.notebook:
      from tqdm.notebook import tqdm
    else:
      from tqdm import tqdm
    self.bar = tqdm(
      desc=f"Epoch {epoch}",
      total=self.length or (len(trainer.dataset) if self.samples else len(trainer.loader)),
      unit='samples' if self.samples else 'batches'
    )

  def on_epoch_end(
    self, epoch=None, loss=None, metrics=None, trainer=None
  ):
    self.bar.close()

  def on_batch_end(
    self,
    batch=None,
    inputs=None,
    targets=None,
    outputs=None,
    loss=None,
    trainer=None
  ):
    if self.samples:
      self.bar.update(len(inputs))
    else:
      self.bar.update(1)



import numpy as np
import torch
from torch.utils.data.dataloader import default_collate

from .utils import pad_to_largest


class PadCollate:
  def __init__(self, pad_inputs=True, pad_targets=False, value=0):
    self.pad_inputs = pad_inputs
    self.pad_targets = pad_targets
    self.value = value

  def __call__(self, batch):
    inputs, targets = zip(*batch)

    if self.pad_inputs:
      inputs = pad_to_largest(list(inputs), self.value)
    if self.pad_targets:
      targets = pad_to_largest(list(targets), self.value)

    return torch.stack(inputs), torch.stack(targets)


class FilterCollate:
  def __init__(
    self, filter=None, collate=default_collate, value=None
  ):
    self.filter = filter or (
      lambda items: [it for it in items if it is not value]
    )
    self.collate = collate

  def __call__(self, batch):
    filtered = self.filter(batch)
    return self.collate(filtered)


filter_collate = FilterCollate()


def image_collate(batch, memory_format=torch.contiguous_format):
  """
  Note: using tensor cores channels_last should lead to better performance
    inspired by: https://github.com/NVIDIA/apex/blob/master/examples/imagenet/main_amp.py#L28
  Args:
    batch: [(image, target), ...]
  """
  images, targets = zip(*batch)
  targets = torch.tensor(targets, dtype=torch.int64)

  w, h = images[0].shape[:2]

  tensor = torch.zeros(
    (len(images), 3, h, w),
    dtype=torch.uint8,
  ).contiguous(memory_format=memory_format)

  for i, img in enumerate(images):
    nump_array = np.asarray(img, dtype=np.uint8)
    if (nump_array.ndim < 3):
      nump_array = np.expand_dims(nump_array, axis=-1)
    nump_array = np.rollaxis(nump_array, 2)
    tensor[i] += torch.from_numpy(nump_array)
  return tensor, targets



from .mixconv import MixConv
from .conv import ConvBlock, ConvBlock1x1, ConvBlock3x3
from .depthwise import DepthwiseConv2d, DepthwiseSeparableConv2d
from .squeeze_excitation import SqueezeExcitation
from .attention import EfficientChannelAttention


import torch

from .. import eval_mode
from .core import inference_stream


class Predictor:
  # model: torch.nn.Module
  def __init__(self, model):
    self.model: torch.nn.Module = model


  def __call__(self, *args, **kwargs):
    x = self.load(*args, **kwargs)
    x = self.transform(x)
    x = self.predict(x)
    x = self.posprocess(x)

    return x

  def export(self):
    pass

  def load(self, samples):
    return samples

  def transform(self, inputs):
    return inputs

  def predict(self, inputs):
    return inputs

  def posprocess(self, inputs):
    return inputs

  def stream(self, batches):
    pass


class Classifier(Predictor):
  def __init__(self, model, classes=None, preprocess=None, postprocess=None):
    self.model: torch.nn.Module = model
    self.preprocess = preprocess
    self.postprocess = postprocess

    self.classes = classes

  def __call__(self, *args, **kwargs):
    x = self.preprocess(*args, **kwargs)
    with eval_mode(self.model):
      x = self.model(x)

    return self.classes.decode(x)



from pyarrow import parquet as pq
import pyarrow as pa
import pandas as pd


def write_parquet(dest, data, columns=None, **kwargs):
  if isinstance(data, dict):
    data = pd.DataFrame(data)

  if isinstance(data, pd.DataFrame):
    return data.to_parquet(dest, **kwargs)
  else:
    d = next(data)
    df = pd.DataFrame(d)
    table = pa.Table.from_pandas(df)
    with pq.ParquetWriter(dest, schema=table.schema, **kwargs) as writer:
      writer.write_table(table)
      for d in next(data):
        writer.write_table(pa.Table.from_pandas(pd.DataFrame(d))


def read_parquet():
  pass


from torch import nn

# TODO: handle downsampling logic

def residual(input, block, identity=None):
  p = block(input)
  if identity:
    input = identity(input)
  return input + p


class Residual(nn.Module):
  def __init__(self, block, identity=None):
    super().__init__()
    self.block = block
    self.identity = identity

  def forward(self, input):
    processed = self.block(input)
    if self.identity:
      input = self.identity(input)
    return input + processed



from torch import nn
from ..stack import Stack


class ConvBlock(Stack):

  class default:
    conv = nn.Conv2d
    norm = nn.BatchNorm2d
    activation = nn.ReLU

  def __init__(
    self,
    in_channels=None,
    out_channels=None,
    kernel_size=None,
    stride=1,
    padding=0,
    dilation=1,
    groups=1,
    bias=None,
    padding_mode='zeros',
    conv=True,
    norm=None,
    activation=True,
    order=('conv', 'norm', 'activation'),
    **extra
  ):
    if conv is True:
      if bias is None:
        # disable conv bias if norm layer is used
        bias = norm is False
      conv = self.default.conv(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        bias=bias,
        stride=stride,
        padding=padding,
        dilation=dilation,
        groups=groups,
        padding_mode=padding_mode
      )
    if norm is True:
      if conv is not None:
        norm = self.default.norm(
          num_features=conv.in_channels
          if order.index('conv') > order.index('norm') else conv.out_channels
        )
      else:
        norm = self.default.norm(in_channels)

    if activation is True:
      activation = self.default.activation() if isinstance(
        self.default.activation, type
      ) else self.default.activation

    d = dict(**extra, conv=conv, norm=norm, activation=activation)
    # NOTE: this depends on dict ordering
    super(ConvBlock, self).__init__(**{k: d[k] for k in order})




class ConvBlock1x1(ConvBlock):
  def __init__(self, **kwargs):
    if kwargs.get('kernel_size', 1) != 1:
      raise ValueError(f'kernel_size must be `1` if provided as an argument, got kernel_size={kwargs["kernel_size"]}')
    super(ConvBlock1x1, self).__init__(kernel_size=1, padding=0, **kwargs)


class ConvBlock3x3(ConvBlock):
  def __init__(self, **kwargs):
    if kwargs.get('kernel_size', 3) != 3:
      raise ValueError(f'kernel_size must be `3` if provided as an argument, got kernel_size={kwargs["kernel_size"]}')
    super(ConvBlock3x3, self).__init__(kernel_size=3, padding=1, **kwargs)


class Viewer:
  def __init__(self, data):
    self.data = data
    self.renderers = {}

  def show(self):
    pass

  def __getitem__(self):
    pass


class PandasViewer(Viewer):
  pass


class DatasetViewer(Viewer):
  pass


from pathlib import Path
import logging
from torch.utils.tensorboard import SummaryWriter
from torch import nn

from .base import Callback


log = logging.getLogger(__name__)

class Tensorboard(Callback):
  writer: SummaryWriter
  def __init__(self, root=None, trainer=None, writer=None):
    self.root = root
    self.trainer = trainer
    self.writer = writer

    self.logged_model_graph = False

  def _init_writer(self):
    self.writer = SummaryWriter(self.root)

  def log(self, val, tag=None, step=None, **kwargs):
    if step is None and self.trainer:
      step = self.trainer.num_steps

    if isinstance(val, nn.Module):
      self.writer.add_graph(
        val,
        kwargs.get('input'),
        verbose=kwargs.get('verbose', False)
      )

  def show(self, root=None):
    from IPython import get_ipython
    ipython = get_ipython()
    ipython.magic('load_ext tensorboard')

    if isinstance(self, (str, Path)):
      # static method (Tensorboard.show())
      ipython.magic(f'tensorboard --logdir {self}')
    else:
      ipython.magic(f'tensorboard --logdir {root or self.root or "./"}')

  def close(self):
    if self.writer:
      self.writer.close()

  def flush(self):
    if self.writer:
      self.writer.flush()

  def on_train_start(self, trainer=None, **kwargs):
    if self.root is None:
      self.root = trainer.paths.tensorboard
    self.trainer = trainer

    if self.writer is None:
      self._init_writer()

  def on_train_end(self, trainer=None):
    if self.trainer and self.trainer.params and self.trainer.history.val_metrics:
      try:
        self.writer.add_hparams(
          dict(self.trainer.params),
          self.trainer.history.val_metrics.summary()
        )
      except ValueError as e:
        log.error(e)

  def on_batch_start(self, inputs=None, **kwargs):
    if not self.logged_model_graph:
      self.writer.add_graph(self.trainer.model, inputs)
      self.logged_model_graph = True

  def on_batch_end(
    self,
    batch=None,
    inputs=None,
    targets=None,
    outputs=None,
    loss=None,
    trainer=None
  ):
    self.writer.add_scalar('train/loss', loss, global_step=trainer.num_steps)
    for metric, values in trainer.history.metrics.items():
      self.writer.add_scalar(f'train/{metric}', values[-1], global_step=len(values) - 1)

  def on_validation_end(self, targets=None, outputs=None, loss=None, trainer=None):
    for metric, values in trainer.history.val_metrics.items():
      self.writer.add_scalar(f'validation/{metric}', values[-1], global_step=len(values) - 1)


from torch.utils.data import DataLoader as TorchLoader, RandomSampler, SequentialSampler, Dataset
from typing import Union, Iterable, Optional, Callable

from ..datasets import TransformDataset
import yann


class DataLoader(TorchLoader):
  """
  Reuse the same iterator for multiple epochs to avoid startup penalty of
  initializing it each time

  (inspired by loader from `timm`)

  # might be fixed here https://github.com/pytorch/pytorch/pull/35795
  """
  def __init__(self, *args, **kwargs):
    super(DataLoader, self).__init__(*args, **kwargs)
    self.__initialized = False
    self.batch_sampler = LoopSampler(self.batch_sampler)
    self.__initialized = True
    self.iterator = super().__iter__()

  def __len__(self):
    return len(self.batch_sampler.sampler)

  def __iter__(self):
    for i in range(len(self)):
      yield next(self.iterator)


class LoopSampler:
  def __init__(self, sampler):
    self.sampler = sampler

  def __iter__(self):
    while True:
      yield from self.sampler


class TransformLoader(DataLoader):
  def __init__(self, dataset, transform, **kwargs):
    super(TransformLoader, self
         ).__init__(TransformDataset(dataset, transform), **kwargs)


def loader(
  data: Union[str, Iterable, Dataset, TorchLoader],
  transform: Optional[Callable] = None,
  **kwargs
):
  """instantiate a loader from a dataset name, dataset or loader"""
  if isinstance(data, DataLoader):
    return data
  if isinstance(data, str):
    data = yann.resolve.dataset(data)
  if isinstance(data, (Dataset, Iterable)):
    if transform:
      return TransformLoader(data, transform=transform, **kwargs)
    else:
      return DataLoader(data, **kwargs)



"""


# TODO:
- add a way to bind params to a function
  # inspired by https://github.com/google/gin-config
  p = Params()

  @p.bind(('batch_size'), prefix=True)
  def train(batch_size: int = 32):
    pass

  @p.bind({'batch_size': 'bs'}, prefix='train')
  def train(batch_size):
    pass

  @p.bind('smooth')
  def loss(x, y, smooth=.2):
    pass

  p.from_command()
  train()

- add __instance_fields__, __fields__ is a class attribute


- Params.sample()
- Params.sampler() # sample without replacement
- Params.grid()

"""

from abc import ABCMeta
from collections import OrderedDict
from copy import deepcopy
from .utils import get_arg_parser


class Field:
  def __init__(self, help=None, type=None, required=False, default=None):
    self.help = help
    self.type = type
    self.required = required
    self.default = default

  def validate(self, val):
    if self.type:
      assert isinstance(val, self.type)

  def __repr__(self):
    return f"{self.__class__.__name__}(type={self.type}, default={self.default})"

  def __str__(self):
    return f"{self.__class__.__name__}(type={self.type}, default={self.default})"


class Choice(Field):
  def __init__(self, choices=None, **kwargs):
    super().__init__(**kwargs)
    self.choices = choices

  def validate(self, val):
    super().validate(val)
    assert val in self.choices


class Range(Field):
  def __init__(self, start, end):
    super(Range, self).__init__()
    self.min = min(start, end)
    self.max = max(start, end)

  def validate(self, val):
    super().validate(val)
    assert self.min <= val <= self.max


class HyperParamsBase:
  def __init__(self, **args):
    self._change_callbacks = []

    for k, v in args.items():
      if k in self.__fields__:
        setattr(self, k, v)
      else:
        raise ValueError(f'Unknown parameter: {k}, should be one of {", ".join(self.__fields__)}')

  def validate(self):
    for k, f in self.__fields__.items():
      try:
        f.validate(getattr(self, k))
      except Exception as e:
        raise Exception(f"{k} failed validation. {e}")

  @classmethod
  def from_command(cls, cmd=None, validate=False, **kwargs):
    parser = get_arg_parser(cls.__fields__, **kwargs)
    parsed = parser.parse_args(cmd.split() if isinstance(cmd, str) else cmd)
    params = cls(**vars(parsed))

    if validate:
      params.validate()

    return params

  @classmethod
  def from_env(cls, prefix=''):
    raise NotImplementedError()

  @classmethod
  def load(cls, path):

    if path.endswith(('yaml', 'yml')):
      import yaml
      with open(path, 'r') as f:
        data = yaml.safe_load(f)
        return cls(**data)
    elif path.endswith('json'):
      import json
      data = json.load(path)
      return cls(**data)

  def save(self, path):
    from .data.io import save_json
    save_json(dict(self), path)



  def on_change(self, callback):
    self._change_callbacks.append(callback)

  def __setattr__(self, k, v):
    if k in self.__fields__:
      for c in self._change_callbacks:
        c(k, v)
    super.__setattr__(self, k, v)

  def __getitem__(self, item):
    if isinstance(item, (tuple, list)):
      return tuple(getattr(self, k) for k in item)
    return getattr(self, item)

  def __setitem__(self, k, v):
    setattr(self, k, v)

  def fork(self, **args):
    return HyperParams(**{**self.items(), **args})

  def __repr__(self):
    return (
      f'{self.__class__.__name__}('
      f"{', '.join(f'{k}={v}' for k, v in self.items())}"
      ')'
    )

  def __str__(self):
    return (
        f'{self.__class__.__name__}(\n' +
        ',\n'.join('  {}={}'.format(k, v) for k, v in self.items()) +
        '\n)'
    )

  def __len__(self):
    return len(self.keys())

  def __hash__(self):
    return hash(tuple(sorted(self.items())))

  def keys(self):
    return self.__fields__.keys()

  def values(self):
    return {getattr(self, k) for k in self.keys()}

  def items(self):
    return ((k, getattr(self, k)) for k in self.keys())

  def inject(self, scope=None, uppercase=True):
    scope = globals() if scope is None else scope
    for k, v in self.items():
      scope[k.upper() if uppercase else k] = v

  @classmethod
  def collect(cls, scope=None, types=(int, str, float, bool),
              upper_only=True, lowercase=True):
    scope = globals() if scope is None else scope

    d = {}
    for k, v in scope.items():
      if types and not isinstance(v, types):
        continue
      if upper_only and not k.isupper():
        continue

      d[k.lower() if lowercase else k] = v

    return cls(**d)

  def __eq__(self, other):
    return hash(self) == hash(other)


class MetaHyperParams(ABCMeta):
  def __new__(metaclass, class_name, bases, namespace):
    fields = OrderedDict()

    for base in reversed(bases):
      if issubclass(base, HyperParamsBase) and base != HyperParamsBase:
        fields.update(deepcopy(base.__fields__))

    existing_attributes = set(dir(HyperParamsBase)) | set(fields)

    new_attributes = {
      k: v for (k, v) in namespace.items()
      if k not in existing_attributes
         and not k.startswith('_')
         and not callable(v)
    }

    for name, annotation in namespace.get('__annotations__', {}).items():
      if name not in new_attributes:
        continue
      if isinstance(annotation, Field):
        fields[name] = annotation
      else:
        fields[name] = Field(type=annotation)

    for name, value in new_attributes.items():
      if name in fields:
        fields[name].default = value
      else:
        fields[name] = Field(default=value, type=type(value))

    return super().__new__(metaclass, class_name, bases, {
      '__fields__': fields,
      **namespace,
    })


class HyperParams(HyperParamsBase, metaclass=MetaHyperParams):
  pass


def sample(params, fields=None):
  pass


def grid(params, fields=None):
  pass





from yann.train import Trainer
import yann
from yann.params import HyperParams
from torch.optim import SGD


class RecognitionParams(HyperParams):
  checkpoint: str = None
  pretrained = True

  warmup = None
  optimizer = 'SGD'
  learning_rate = 0.01
  momentum = .9
  weight_decay = 0.0001


  label_smoothing = 0.1

  loss = 'soft_target_cross_entry'

  notebook = False


  sizes = None  # 160, 160, 224, 224, 360

  mixup = False
  manifold_mixup = False
  cutmix = False
  cutout = False
  autoaugment = False
  augmix = False



class RecognitionTrainer(Trainer):
  def __init__(
      self,
      model=None,
      dataset=None,
      optimizer=None,
      loss=None,
      loader=None,
      sampler=None,
      num_workers=8,
      transform=None,
      transform_batch=None,
      lr_scheduler=None,
      lr_batch_step=False,
      callbacks=None,
      device=None,
      parameters='trainable',
      batch_size=16,
      val_dataset=None,
      val_loader=None,
      val_transform=None,
      classes=None,
      parallel=False,
      name=None,
      description=None,
      root='./experiments/train-runs/',
      metrics=None,
      collate=None,
      params=None,
      pin_memory=True,
      step=None,
      id=None

  ):

    model = model or yann.resolve.model('resnet50', pretrained=True)

    optimizer = optimizer or SGD(model.parameters(), momentum=.9)

    super(RecognitionTrainer, self).__init__(
      model=model or 'resnet50',
      optimizer=optimizer or
    )


from time import time as get_time


def padded_insert(items, index, value, null_val=None):
  """
  insert value into the items list at given index
  if index is larger than length of list then
  extend it up to index and pad the extra space with null_val
  """
  if len(items) == index:
    items.append(value)
  elif len(items) > index:
    items[index] = value
  else:
    items.extend([null_val] * (index - len(items)))
    items.append(value)
  return items


class PaddedList(list):
  def __init__(self, *args, null_val=None):
    super().__init__(*args)
    self.null_val = null_val

  def __setitem__(self, index, value):
    if len(self) == index:
      self.append(value)
    elif len(self) > index:
      super().__setitem__(index, value)
    else:
      self.extend([self.null_val] * (index - len(self)))
      self.append(value)


class MetricStore:
  def __init__(self, names=None, null_val=None, cast_value=None):
    self.values = {n: PaddedList(null_val=null_val) for n in names} if names else {}
    self.times = PaddedList(null_val=null_val)

    self.null_val = null_val
    self.cast_value = cast_value

  def update(self, step='next', time='now', **values):
    step = len(self.times) if step == 'next' else step
    time = get_time() if time == 'now' else time

    if self.cast_value:
      values = {k: self.cast_value(v) for (k, v) in values.items()}

    self.times[step] = time
    for metric, v in values.items():
      if metric not in self.values:
        self._init_metric(metric)
      self.values[metric][step] = v

    # need to make sure all metrics have the same length
    for metric, vs in self.values.items():
      if len(vs) < len(self.times):
        vs[len(self.times) - 1] = self.null_val

  def _init_metric(self, name):
    self.values[name] = PaddedList(null_val=self.null_val)

  def __getitem__(self, name):
    return self.values[name]

  def __contains__(self, item):
    return item in self.values

  def keys(self):
    return self.values.keys()

  def items(self):
    return self.values.items()

  def __len__(self):
    return len(self.times)

  def to_pandas(self):
    import pandas as pd
    return pd.DataFrame({'times': self.times, **self.values})

  def plot(self, metrics=None, time=False, clear=False):
    if clear:
      try:
        from IPython.display import clear_output
        clear_output(wait=True)
      except:
        pass

  def summary(self):
    s = {}
    for metric, values in self.values.items():
      try:
        s[f"{metric}_min"] = min(values)
        s[f"{metric}_max"] = max(values)
      except:
        pass
    return s


  def __repr__(self):
    return  f"MetricStore({', '.join(f'{k}=(min={min(v)}, max={max(v)})' for k,v in self.values.items())}, len={len(self)})"


class EventStore(list):
  def add(self, *args, key=None, value=None, step=None, time=None):
    self.append(Event(key=key, value=value, step=step, time=time))


class Event:
  __slots__ = ['key', 'value', 'step', 'time']

  def __init__(self, key=None, value=None, step=None, time=None):
    self.key = key
    self.value = value
    self.step = step
    self.time = time


# import dataclasses
# from typing import Optional, Dict, Tuple
#
# import uuid
# from enum import Enum
#
#
# class ModelFormat(Enum):
#   onnx = 'onnx'
#   torchscript = 'torchscript'
#   torchmodule = 'torchmodule'
#   scikitlearn = 'scikitlearn'
#   python_function = 'python_function'
#
#
# class TabularFormats(Enum):
#   parquet = 'parquet'
#   csv = 'csv'
#   json = 'json'
#
#
#
# @dataclasses.dataclass()
# class Spec:
#   uri: str
#   name: Optional[str]
#   description: Optional[str]
#
#   slug: str  # human readable id, used in urls
#   version: str
#
#   time_created: int
#
#   metadata: Dict
#
#   source: str
#   tags: Tuple[str]
#
#
# @dataclasses.dataclass()
# class Dataset(Spec):
#   fields: Dict
#   type: str
#   params: dict
#
#
#
# @dataclasses.dataclass()
# class Data(Spec):
#   format: str
#   shape: Tuple
#
# class Vector(Data):
#   pass
#
# class Image(Data):
#   pass
#
# class TabularData(Dataset)
#   fields:List
#
# class Image(Data):
#   pass
#
# @dataclasses.dataclass()
# class Operator(Spec):
#   inputs = Tuple
#   outputs = Tuple
#
#
# @dataclasses.dataclass()
# class Model(Operator):
#   experiment: Optional[str]
#   checkpoint: str
#
#   type: str
#
#
# @dataclasses.dataclass()
# class Predictor(Operator):
#   pass
#
#
#
#
# dataset = TabularData(
#   fields=[
#     Image()
#
#   ]
# )
#
#
#
#
#
# dataset = Data(
#   name='imagenet-b',
#   fields=dict(
#     image=Image(),
#
#   )
# )
#
#
#
#
# model_spec = Model(
#   id=uuid.uuid4(),
#   inputs=(
#     Image(),
#   ),
#   outputs=
# )
#
#
# index_spec = Data(
#   id=uuid.uuid4(),
#   name='index',
#   type='annoy',
#   params=dict()
# )
#
#
#



from functools import wraps

class lazy:
  __slots__ = 'method', 'name'

  def __init__(self, method):
    self.method = method
    self.name = method.__name__

  def __get__(self, obj, cls):
    if obj:
      val = self.method(obj)
      setattr(obj, self.name, val)
      return val


class RobustFunction:
  def __init__(self, func, exceptions=Exception, default=None):
    self.func = func
    self.exceptions = exceptions
    self.default = default

  def __call__(self, *args, **kwargs):
    if not self.exceptions:
      return self.func(*args, **kwargs)
    try:
      r = self.func(*args, **kwargs)
    except self.exceptions:
      r = self.default
    return r


def robust(x=None, exceptions=Exception, default=None):
  if callable(x):
    return RobustFunction(func=x, exceptions=exceptions, default=default)
  else:
    def decorator(x):
      return RobustFunction(func=x, exceptions=exceptions, default=default)

    return decorator


class FunctionTracker:
  def __init__(self, func, sanitize=None):
    self.func = func
    self.history = []
    self.sanitize = sanitize

  def __call__(self, *args, **kwargs):
    r = self.func(*args, **kwargs)
    if self.sanitize:
      r = self.sanitize(r)
    self.history.append(r)
    return r


def track(x=None, sanitize=None):
  if callable(x):
    return FunctionTracker(func=x)
  else:
    def decorator(x):
      return FunctionTracker(func=x, sanitize=sanitize)

    return decorator


import torch
import types

from .classes import Classes
from .loaders import TransformLoader
from .transform import Transformer


def get_dataset_name(x):
  while hasattr(x, 'dataset'):
    x = x.dataset

  if hasattr(x, 'name'):
    return x.name

  return x.__class__.__name__


def batches(*tensors, size=32, shuffle=False, order=None):
  if len(tensors) == 1 and isinstance(tensors[0], str):
    # assume a registered dataset name was passed (like batches('MNIST'))
    import yann
    tensors = (yann.resolve.dataset(tensors[0]),)
  if shuffle:
    order = torch.randperm(len(tensors[0]))

  if len(tensors) == 1:
    for i in range(0, len(tensors[0]), size):
      if order is not None:
        indices = order[i:i+size]
        yield tensors[0][indices]
      else:
        yield tensors[0][i:i+size]
  else:
    for i in range(0, len(tensors[0]), size):
      if order is not None:
        indices = order[i:i+size]
        yield tuple(t[indices] for t in tensors)
      else:
        yield tuple(t[i:i+size] for t in tensors)


def unbatch(batches):
  return (x for b in batches for x in b)


def chunk(sequence, size=32):
  if isinstance(sequence, types.GeneratorType):
    batch = []
    for x in sequence:
      batch.append(x)
      if len(batch) == size:
        yield batch
        batch = []
  else:
    for i in range(0, len(sequence), size):
      yield sequence[i:i+size]


def loop(items):
  while True:
    yield from items


def shuffle(*sequences):
  order = torch.randperm(len(sequences[0]))
  return (
     [s[i] for i in order] if isinstance(s, (tuple, list)) else s[order]
     for s in sequences
  )

def flatten(x, out=None, prefix='', sep='.'):
  """
  Flatten nested dict
  """
  out = out if out is not None else {}

  if isinstance(x, dict):
    for k in x:
      flatten(x[k], out=out, prefix=f"{prefix}{sep if prefix else ''}{k}", sep=sep)
  elif isinstance(x, (list, tuple)):
    for k, v in enumerate(x):
      flatten(k, out=out, prefix=f"{prefix}{sep if prefix else ''}{k}", sep=sep)
  else:
    out[prefix] = x

  return out


def print_tree(root, indent=4):
  from pathlib import Path
  root = Path(root)
  print(f'{root}')
  for path in sorted(root.rglob('*')):
    depth = len(path.relative_to(root).parts)
    print(f'{" " * (depth * indent)} {path.name}')


import pytest
import torch.cuda
from torch import nn
from torch.optim import SGD

from yann.callbacks import (
  History, HistoryPlotter, HistoryWriter, Logger, Checkpoint
)
from yann.datasets import TinyDigits
from yann.datasets.wrappers import Slice
from yann.modules import Flatten
from yann.train import Trainer

from torchvision.models.resnet import resnet18

devices = ['cpu', 'cuda'] if torch.cuda.is_available() else ['cpu']


@pytest.mark.slow
@pytest.mark.parametrize('device', devices)
def test_train(tmpdir, device):
  """Sanity check train run"""

  model = nn.Sequential(
    nn.Conv2d(1, 20, 3),
    nn.ReLU(inplace=True),
    nn.Conv2d(20, 20, 3),
    nn.ReLU(inplace=True),
    Flatten(),
    nn.Linear(320, 10)
  )

  train = Trainer(
    root=tmpdir,
    model=model,
    dataset=Slice(TinyDigits(), 0, 256),
    device=device,
    optimizer=SGD(model.parameters(), lr=.01, momentum=0.9, weight_decay=.001),
    loss=nn.CrossEntropyLoss(),
    callbacks=[
      History(),
      HistoryPlotter(save=True),
      HistoryWriter(),
      Logger(batch_freq=20),
      Checkpoint()
    ]
  )

  train(2)

  assert train.paths.checkpoints.is_dir()
  assert train.history.metrics

  export_path = train.export()

  assert export_path
  assert export_path.is_dir()


# @pytest.mark.slow
# @pytest.mark.parametrize('device', devices)
# def test_train_resolved(tmpdir, device):
#   from yann.data.transform import ImageTransformer
#
#   train = Trainer(
#     root=tmpdir,
#     model='densenet121',
#     dataset='CIFAR10',
#     loss='CrossEntropy',
#     optimizer='SGD',
#     transform=ImageTransformer(resize=224)
#   )
#
#   # train(1)


def test_transforms():

  t = Trainer(
    transform={
      'mask': 'foo',
      'label': 'foo'
    }
  )


from torch import nn
from ..stack import Stack
from ..residual import Residual


class SqueezeExcitation(Residual):

  def __init__(self, channels: int, reduction: int):
    """
    Args:
      channels:
      reduction: reduction factor, paper claims 16 was the best value
    """
    inner_channels = channels // reduction
    super().__init__(
      Stack(
        nn.AdaptiveAvgPool2d(1),
        nn.Conv2d(channels, inner_channels, kernel_size=1, padding=0),
        nn.ReLU(inplace=True),
        nn.Conv2d(inner_channels, channels, kernel_size=1, padding=0),
        nn.Sigmoid()
      )
    )


SE = SqueezeExcitation


# from yann.data.io.download import Downloader
#
#
# class RemoteDataset:
#   def __init__(self, downloader: Downloader):
#     self.downloader = downloader
#
#   def prefetch(self, indices=None):
#     for n in indices:
#       self.downloader.enqueue(self[n][0])
#
#   def __getitem__(self, item):
#     return 'id', ['labels']
#
# def test_downloader():
#   downloader = Downloader('.')
#
#   local_path = downloader['gs://foo']
#
#   dataset = RemoteDataset(downloader)
#
#   for uri, labels in dataset:
#     downloader.enqueue(uri)
#
#
#   for key, result in downloader.results:
#     pass
#
#   for key, result in downloader.as_completed():
#     pass
#
#
#   for key, error in downloader.errors:
#     pass
#
#
#   downloader.enqueue('asd')
#
#
#   downloader.on_error(lambda key, error: print(error))
#   downloader.on_success(lambda key, result: print('completed', key))
#
#   downloader.complete()
#
#
#   def unpack(x):
#     pass
#
#
#   dataset_downloader = Downloader('./')
#
#   dataset_downloader.on_success(unpack)
#   dataset_downloader.enqueue('s3://1.tar')
#
#
#   dataset_downloader.get_all(['1', '2', '3'])
#   dataset_downloader.stream(['1', '2', '3'])


import yann
import torch


def test_io(tmpdir):
  x = {
    'a': 2
  }

  yann.save(x, tmpdir / 'x.json')
  y = yann.load(tmpdir / 'x.json')
  assert x == y

  yann.save(x, tmpdir / 'x.th')
  y = yann.load(tmpdir / 'x.th')
  assert x == y

  yann.save(x, tmpdir / 'x.pickle')
  y = yann.load(tmpdir / 'x.pickle')
  assert x == y


from torch.utils.data.sampler import Sampler
import random
from collections import defaultdict

from . import batches


class BalancedTargetSampler(Sampler):
  def __init__(self, dataset, targets=None, size=None):
    super().__init__(dataset)
    self.dataset = dataset
    self.size = size

    self.target_to_indices = defaultdict(list)

    if not targets and hasattr(dataset, 'targets'):
      targets = dataset.targets

    if targets:
      for n, ts in enumerate(targets):
        if isinstance(ts, (list, tuple)):
          for t in ts:
            self.target_to_indices[t].append(n)
        else:
          self.target_to_indices[ts].append(n)
    else:
      for n, (x, ts) in enumerate(dataset):
        if isinstance(ts, (list, tuple)):
          for t in ts:
            self.target_to_indices[t].append(n)
        else:
          self.target_to_indices[ts].append(n)

    self.targets = list(self.target_to_indices)

  def __iter__(self):
    for n in range(len(self)):
      t = random.choice(self.targets)
      yield random.choice(self.target_to_indices[t])

  def __len__(self):
    return self.size or len(self.dataset)


from torch.utils.data import Sampler


class BatchShuffleSampler(Sampler):
  """
  Splits sorted data into `batch_size` chunks, then shuffles those chunks.

  Useful when dataset is sorted by length and want each batch to have items
  of around same length but want to randomize the batch order

  ex:
    list(BatchShuffleSampler(range(12), 3))
    >> [9, 10, 11, 3, 4, 5, 6, 7, 8, 0, 1, 2]
  """

  def __init__(self, data, batch_size):
    super().__init__(data)
    self.data = data
    self.batch_size = batch_size

  def __len__(self):
    return len(self.data)

  def __iter__(self):
    bs = list(batches(range(len(self)), size=self.batch_size))
    random.shuffle(bs)
    return (s for b in bs for s in b)


from .base import Callback


class Checkpoint(Callback):
  def __init__(self, freq=1):
    self.freq = freq
    self.paths = {}

  def on_epoch_end(self, epoch, loss=None, metrics=None, trainer=None):
    if epoch % self.freq == 0 :
      self.paths[epoch] = trainer.checkpoint()




from logging import Logger as PythonLogger


class Logger(PythonLogger):
  def __init__(self, **context):
    super(Logger, self).__init__()

    self.context = context

  def log(self, level: int, msg: Any, *args: Any, exc_info: _ExcInfoType = ...,
                stack_info: bool = ..., extra: Optional[Dict[str, Any]] = ...,
                **kwargs: Any) -> None:
    pass

  def __call__(self, *args, **kwargs):
    pass


#
# log = Logger()
#
#
# import torch
# log.debug(tensor=torch.rand(3,3))
# log.info(image=image)
# log.info()
#
#
# log.warning()


from contextlib import contextmanager

__version__ = '0.0.40'

import torch
from torch import nn
import numpy as np
from .config.setup import registry

register = registry
resolve = registry.resolve

from pathlib import Path

from .utils import to_numpy, repeat, counter


from .data import batches, shuffle, chunk
from .data.io import load, save
from .data.io.download import download
from .data.utils import pad, pad_to_largest
# from .data import datasets
from .viz import show, plot
from .utils.timer import time
from .utils.profile import profile

from .testing import Checker
from .data.loaders import loader

# T = torch.Tensor
#
# check = Checker()
#
#
#
# def serialize():
#   pass
#
# def deserialize():
#   pass
#
#
# class Cache:
#   def __call__(self, data):
#     pass
#
# def cache(data=None, path=None, recompute=None, validate=False, hash=True):
#   """
#
#   processed_data
#
#   yann.cache(processed_data
#
#
#   Args:
#     data:
#     path:
#     recompute:
#     validate:
#     hash:
#
#   Returns:
#
#   """
#
#
# def checkpoint():
#   pass
#
# def load_checkpoint():
#   pass


class default:
  root = Path('~/.yann/')

  device = torch.device('cuda') \
    if torch.cuda.is_available() else torch.device('cpu')

  batch_size = None
  num_workers = None
  optimizer = None

  callbacks = None

  train_root = None
  datasets_root = root / 'datasets'

  checkpoint_name_format = ''


def seed(val=1, deterministic=False):
  import numpy as np
  import torch
  import random
  random.seed(val)
  np.random.seed(val)
  torch.manual_seed(val)
  try:
    torch.cuda.manual_seed(val)

    if deterministic:
      torch.cuda.deterministic = True
      # torch.cuda.benchmark = False
  except:
    pass
  return val


def get_item(x):
  if torch.is_tensor(x):
    if x.is_cuda:
      x = x.cpu()
    return x.item()
  elif isinstance(x, np.ndarray):
    return x.item()
  return x


def benchmark():
  from torch.backends import cudnn
  cudnn.benchmark = True


def detect_anomalies(val=True):
  import torch.autograd
  torch.autograd.set_detect_anomaly(val)


def evaluate(model, batches, device=None):
  for x, y in batches:
    if device:
      x, y = x.to(device), y.to(device)

    model.eval()
    with torch.no_grad():
      pred = model(x)

    yield x, y, pred


def predict_multicrop(model, inputs, reduce='mean'):
  batch_size, num_crops, *sample_shape = inputs.shape
  flat_preds = model(inputs.view(-1, *sample_shape))
  outputs = flat_preds.view(batch_size, num_crops, -1)
  if reduce:
    outputs = getattr(torch, reduce)(outputs, 1)
  return outputs


class Multicrop(torch.nn.Module):
  def __init__(self, model):
    super().__init__()
    self.model = model

  def forward(self, inputs, *rest):
    return predict_multicrop(self.model, inputs)


def set_param(x, param, val):
  if hasattr(x, 'param_groups'):
    for group in x.param_groups:
      group[param] = val
  else:
    setattr(x, param, val)


def scale_param(x, param, mult):
  if hasattr(x, 'param_groups'):
    for group in x.param_groups:
      group[param] *= mult
  else:
    setattr(x, param, getattr(x, param) * mult)


def group_params(model, get_key):
  splits = {}
  for name, param in model.named_parameters():
    splits[get_key(name, param)] = param
  return splits

from torch.nn.modules.batchnorm import _BatchNorm


def split_regularization_params(
    module: nn.Module,
    excluded_modules=(_BatchNorm,),
    excluded_names=('bias',),
    param_groups=True,
):
  """
  filter out parameters which should not be regularized
  """
  reg, no_reg = [], []
  m: nn.Module
  for m in module.modules():
    if isinstance(m, excluded_modules):
      no_reg.extend((x for x in m.parameters(recurse=False) if x is not None))
    else:
      for name, param in m.named_parameters(recurse=False):
        if param is not None:
          if name.endswith(excluded_names):
            no_reg.append(param)
          else:
            reg.append(param)
  if param_groups:
    return [dict(params=reg), dict(params=no_reg, weight_decay=0)]
  else:
    return reg, no_reg


def trainable(parameters):
  return (p for p in parameters if p.requires_grad)


# TODO: handle batchnorm
def freeze(x, exclude=None):
  if isinstance(x, nn.Module):
    if exclude:
      for m in x.modules():
        if not isinstance(m, exclude):
          for p in m.parameters(recurse=False):
            p.requires_grad = False
    else:
      for p in x.parameters():
        p.requires_grad = False
  elif exclude:
    raise ValueError(
      "can't exclude modules if parameters are passed, "
      "pass an instance of nn.Module if you need to "
      "exclude certain modules")
  else:
    for p in x:
      p.requires_grad = False

  return x


def freeze_non_batchnorm(x):
  """
  Freeze layers except batchnorm
  should be used when fine tuning / transfer learning
  """
  return freeze(x, exclude=_BatchNorm)


def unfreeze(parameters):
  if isinstance(parameters, nn.Module):
    parameters = parameters.parameters()
  for p in parameters:
    p.requires_grad = True


def filter_modules(module: nn.Module, type, named=True):
  for n, m in module.named_modules():
    if isinstance(m, type):
      if named:
        yield n, m
      else:
        yield m


def replace_linear(model, num_outputs, layer_name=None):
  if layer_name is None:
    linear_layers = list(filter_modules(model, nn.Linear, named=True))
    if len(linear_layers) == 1:
      layer_name = linear_layers[0][0]
    elif len(linear_layers) == 0:
      raise ValueError('No linear layers found in model')
    else:
      raise ValueError(
        f'Multiple linear layers found and layer name was not provided, '
        f'provide a valid layer_name, '
        f'(valid names: {", ".join([n for n, m in linear_layers])})'
      )


  if '.' in layer_name:
    *path, layer_name = list(layer_name.split('.'))
    for p in path:
      model = getattr(model, p)

  old_linear = getattr(model, layer_name)
  new_linear = nn.Linear(old_linear.in_features, num_outputs)
  new_linear.to(old_linear.weight.device)

  setattr(
    model,
    layer_name,
    new_linear
  )

  return layer_name


def to_fp16(model):
  # https://discuss.pytorch.org/t/training-with-half-precision/11815
  # https://github.com/csarofeen/examples/tree/fp16_examples_cuDNN-ATen/imagenet
  # https://github.com/NVIDIA/apex
  # https://github.com/NVIDIA/apex/tree/master/apex/amp
  model.half()
  for layer in model.modules():
    if isinstance(layer, torch.nn.BatchNorm2d):
      layer.float()


def none_grad(model: nn.Module):
  """
  more efficient version of zero_grad()
  """
  for p in model.parameters():
    p.grad = None


@contextmanager
def eval_mode(*modules, grad=False):
  if grad:
    training = (m.training for m in modules)
    try:
      for m in modules: m.eval()
      yield
    finally:
      for m, train in zip(modules, training):
        if train:
          m.train()
  else:
    with torch.no_grad():
      training = (m.training for m in modules)
      try:
        for m in modules: m.eval()
        yield
      finally:
        for m, train in zip(modules, training):
          if train:
            m.train()


@contextmanager
def train_mode(*modules):
  initial_training_states = (m.training for m in modules)
  try:
    for m in modules:
      m.train()

    yield

  finally:
    for m, train in zip(modules, initial_training_states):
      if train:
        m.train()
      else:
        m.eval()


@contextmanager
def optim_step(optimizer, zero_grad=True):
  if zero_grad:
    optimizer.zero_grad()

  yield

  optimizer.step()


def to(*items, **kwargs):
  """call `.to()` on all items that have a `to()` method, skips ones that don't"""
  return tuple(x.to(**kwargs) if hasattr(x, 'to') else x for x in items)



from torchvision.datasets.folder import ImageFolder as IF, default_loader

from yann.data import Classes


class ImageFolder(IF):
  def __init__(
    self,
    root,
    transform=None,
    target_transform=None,
    loader=default_loader,
    is_valid_file=None
  ):
    super(ImageFolder, self).__init__(
      root=root,
      transform=transform,
      target_transform=target_transform,
      loader=loader,
      is_valid_file=is_valid_file
    )

    self.classes = Classes(self.classes)


import random

adjectives = [
  "able",
  "about",
  "above",
  "abuzz",
  "ace",
  "achy",
  "acid",
  "acned",
  "acute",
  "adept",
  "adorable",
  "adult",
  "adventurous",
  "afire",
  "afoot",
  "afoul",
  "aft",
  "after",
  "aged",
  "aggressive",
  "agile",
  "aging",
  "aglow",
  "ago",
  "agreeable",
  "ahead",
  "aided",
  "airy",
  "ajar",
  "akin",
  "alert",
  "alien",
  "alike",
  "alive",
  "all",
  "alone",
  "aloof",
  "alpha",
  "alto",
  "amber",
  "ample",
  "amused",
  "angry",
  "annoyed",
  "annoying",
  "anti",
  "antic",
  "antsy",
  "anxious",
  "any",
  "apart",
  "apish",
  "apt",
  "arced",
  "arch",
  "arid",
  "armed",
  "arrogant",
  "ashamed",
  "ashen",
  "ashy",
  "askew",
  "astir",
  "atrip",
  "attic",
  "attractive",
  "average",
  "avian",
  "avid",
  "awake",
  "aware",
  "awash",
  "away",
  "awed",
  "awful",
  "awing",
  "awned",
  "awry",
  "axial",
  "azure",
  "back",
  "bad",
  "baggy",
  "baked",
  "bald",
  "balmy",
  "bandy",
  "bare",
  "bared",
  "basal",
  "base",
  "based",
  "basic",
  "bated",
  "bats",
  "batty",
  "bay",
  "beady",
  "beamy",
  "beat",
  "beautiful",
  "beefy",
  "beery",
  "beige",
  "bent",
  "best",
  "beta",
  "better",
  "bewildered",
  "bias",
  "birch",
  "bitty",
  "black",
  "blame",
  "bland",
  "blank",
  "bleak",
  "blear",
  "blind",
  "blond",
  "bloody",
  "blown",
  "blue",
  "blue-eyed",
  "bluff",
  "blunt",
  "blushing",
  "boggy",
  "bogus",
  "bold",
  "bone",
  "boned",
  "bonny",
  "bony",
  "boon",
  "boozy",
  "bored",
  "born",
  "boss",
  "bossy",
  "both",
  "bound",
  "bowed",
  "boxed",
  "boxy",
  "brag",
  "brainy",
  "brash",
  "brave",
  "breakable",
  "brief",
  "bright",
  "briny",
  "brisk",
  "broad",
  "broke",
  "brown",
  "brute",
  "buff",
  "buggy",
  "built",
  "bulgy",
  "bulky",
  "bully",
  "bum",
  "bumpy",
  "burly",
  "burnt",
  "bush",
  "bushy",
  "bust",
  "busty",
  "busy",
  "butch",
  "calm",
  "camp",
  "campy",
  "careful",
  "catty",
  "cautious",
  "charming",
  "cheap",
  "cheerful",
  "chewy",
  "chic",
  "chief",
  "civic",
  "civil",
  "clean",
  "clear",
  "cleft",
  "clever",
  "close",
  "cloudy",
  "clumsy",
  "cocky",
  "cod",
  "cold",
  "color",
  "colorful",
  "combative",
  "comfortable",
  "comfy",
  "comic",
  "concerned",
  "condemned",
  "confused",
  "cool",
  "cooperative",
  "coral",
  "corny",
  "cosy",
  "courageous",
  "coy",
  "cozy",
  "crazy",
  "creepy",
  "crisp",
  "cross",
  "crowded",
  "cruel",
  "cubic",
  "cured",
  "curious",
  "curly",
  "curt",
  "curvy",
  "cushy",
  "cut",
  "cute",
  "cyan",
  "daft",
  "daily",
  "damn",
  "damp",
  "dandy",
  "dangerous",
  "dank",
  "dark",
  "dated",
  "dazed",
  "dead",
  "deaf",
  "dear",
  "deep",
  "defeated",
  "defiant",
  "deft",
  "deist",
  "delightful",
  "dense",
  "depressed",
  "determined",
  "dewy",
  "dicey",
  "different",
  "difficult",
  "dim",
  "dingy",
  "dinky",
  "dire",
  "dirt",
  "dirty",
  "disgusted",
  "distinct",
  "disturbed",
  "dizzy",
  "dodgy",
  "domed",
  "done",
  "doped",
  "dopey",
  "dopy",
  "dormy",
  "dosed",
  "doubtful",
  "down",
  "downy",
  "dozen",
  "drab",
  "drawn",
  "dread",
  "drear",
  "dress",
  "dried",
  "droll",
  "drunk",
  "dry",
  "dual",
  "dud",
  "due",
  "dull",
  "dumb",
  "dummy",
  "dusky",
  "dusty",
  "dyed",
  "dying",
  "each",
  "eager",
  "early",
  "eased",
  "east",
  "easy",
  "edged",
  "edgy",
  "eerie",
  "eight",
  "elated",
  "elder",
  "elect",
  "elegant",
  "elfin",
  "elite",
  "embarrassed",
  "empty",
  "enchanting",
  "encouraging",
  "ended",
  "energetic",
  "enthusiastic",
  "envious",
  "epic",
  "equal",
  "even",
  "every",
  "evil",
  "exact",
  "excited",
  "expensive",
  "extra",
  "exuberant",
  "eyed",
  "fab",
  "faced",
  "faded",
  "faint",
  "fair",
  "faithful",
  "fake",
  "false",
  "famed",
  "famous",
  "fancy",
  "fantastic",
  "far",
  "fast",
  "fat",
  "fatal",
  "fated",
  "fazed",
  "feral",
  "few",
  "fewer",
  "fierce",
  "fiery",
  "fifth",
  "fifty",
  "filmy",
  "filthy",
  "final",
  "fine",
  "finer",
  "fired",
  "firm",
  "first",
  "fishy",
  "fit",
  "five",
  "fixed",
  "fizzy",
  "flaky",
  "flash",
  "flat",
  "fleet",
  "flint",
  "flip",
  "fluid",
  "flush",
  "fly",
  "foamy",
  "focal",
  "foggy",
  "fond",
  "foolish",
  "fore",
  "foul",
  "found",
  "four",
  "foxy",
  "fragile",
  "frail",
  "frank",
  "frantic",
  "free",
  "fresh",
  "fried",
  "friendly",
  "frightened",
  "front",
  "full",
  "fumed",
  "funky",
  "funny",
  "furry",
  "fused",
  "fussy",
  "fuzzy",
  "game",
  "gaudy",
  "gaunt",
  "gawky",
  "gentle",
  "giant",
  "giddy",
  "gifted",
  "gimpy",
  "glad",
  "glamorous",
  "gleaming",
  "glorious",
  "glum",
  "godly",
  "going",
  "gold",
  "gone",
  "good",
  "gooey",
  "goofy",
  "gorgeous",
  "gory",
  "graceful",
  "grand",
  "great",
  "green",
  "grey",
  "grieving",
  "grim",
  "grimy",
  "gross",
  "grotesque",
  "grown",
  "gruff",
  "grumpy",
  "gummy",
  "gushy",
  "gusty",
  "gutsy",
  "hairy",
  "hale",
  "half",
  "halt",
  "hammy",
  "handsome",
  "handy",
  "happy",
  "hard",
  "hardy",
  "harsh",
  "hasty",
  "hated",
  "hazel",
  "hazy",
  "healthy",
  "heard",
  "heavy",
  "hefty",
  "held",
  "helpful",
  "helpless",
  "here",
  "hex",
  "hexed",
  "high",
  "hilarious",
  "hilly",
  "hind",
  "hip",
  "hired",
  "hoar",
  "hoary",
  "hokey",
  "holey",
  "holy",
  "home",
  "homeless",
  "homely",
  "homey",
  "honey",
  "horny",
  "horrible",
  "hot",
  "huffy",
  "huge",
  "human",
  "humid",
  "hungry",
  "hurt",
  "husky",
  "icky",
  "icy",
  "ideal",
  "idle",
  "iffy",
  "ill",
  "important",
  "impossible",
  "inane",
  "inept",
  "inert",
  "inexpensive",
  "inky",
  "inner",
  "innocent",
  "inquisitive",
  "ionic",
  "irate",
  "iron",
  "itchy",
  "jade",
  "jaded",
  "jaggy",
  "jawed",
  "jazzy",
  "jealous",
  "jet",
  "jittery",
  "joint",
  "jolly",
  "jowly",
  "joyous",
  "juicy",
  "jumbo",
  "jumpy",
  "just",
  "kempt",
  "key",
  "keyed",
  "khaki",
  "kin",
  "kind",
  "kinky",
  "known",
  "kooky",
  "laced",
  "lacy",
  "laid",
  "lame",
  "lank",
  "lanky",
  "large",
  "last",
  "late",
  "later",
  "lax",
  "lay",
  "lazy",
  "leafy",
  "leaky",
  "lean",
  "least",
  "left",
  "legal",
  "less",
  "level",
  "light",
  "like",
  "liked",
  "limp",
  "lined",
  "lit",
  "live",
  "lively",
  "liver",
  "livid",
  "loamy",
  "local",
  "loco",
  "lofty",
  "lone",
  "lonely",
  "long",
  "loony",
  "loopy",
  "loose",
  "lossy",
  "lost",
  "loud",
  "lousy",
  "loved",
  "lovely",
  "low",
  "lowly",
  "loyal",
  "lucid",
  "lucky",
  "lumpy",
  "lunar",
  "lurid",
  "lush",
  "lusty",
  "lyric",
  "macho",
  "macro",
  "mad",
  "made",
  "magic",
  "magnificent",
  "main",
  "major",
  "male",
  "mangy",
  "manic",
  "manly",
  "many",
  "mass",
  "matt",
  "matte",
  "mauve",
  "mealy",
  "mean",
  "meaty",
  "meek",
  "meet",
  "mere",
  "merry",
  "messy",
  "metal",
  "micro",
  "mild",
  "milky",
  "mimic",
  "mined",
  "mini",
  "minor",
  "mint",
  "minty",
  "minus",
  "mired",
  "mirky",
  "misty",
  "mixed",
  "mock",
  "mod",
  "modal",
  "model",
  "modern",
  "moist",
  "molar",
  "moldy",
  "mono",
  "moody",
  "moony",
  "moot",
  "moral",
  "more",
  "mossy",
  "most",
  "mothy",
  "motionless",
  "motor",
  "mousy",
  "moved",
  "mown",
  "much",
  "mucky",
  "muddy",
  "muggy",
  "mum",
  "mural",
  "murky",
  "mushy",
  "musky",
  "must",
  "musty",
  "mute",
  "muted",
  "mysterious",
  "naive",
  "nary",
  "nasal",
  "nasty",
  "natal",
  "natty",
  "naughty",
  "naval",
  "near",
  "neat",
  "needy",
  "nervous",
  "nervy",
  "new",
  "newsy",
  "next",
  "nice",
  "nifty",
  "nigh",
  "nine",
  "ninth",
  "noble",
  "noisy",
  "none",
  "north",
  "nosed",
  "noted",
  "novel",
  "nubby",
  "numb",
  "nuts",
  "nutty",
  "oaken",
  "oaten",
  "obedient",
  "obese",
  "obnoxious",
  "odd",
  "oiled",
  "oily",
  "okay",
  "old",
  "old-fashioned",
  "olden",
  "older",
  "olive",
  "one",
  "only",
  "oozy",
  "open",
  "optic",
  "oral",
  "other",
  "out",
  "outer",
  "outrageous",
  "outstanding",
  "oval",
  "over",
  "overt",
  "owing",
  "own",
  "owned",
  "pagan",
  "paid",
  "pale",
  "palmy",
  "panicky",
  "pass",
  "past",
  "pasty",
  "pat",
  "paved",
  "peaky",
  "peaty",
  "pedal",
  "pent",
  "peppy",
  "perfect",
  "perky",
  "pert",
  "pesky",
  "pet",
  "petty",
  "phony",
  "piano",
  "picky",
  "pied",
  "piggy",
  "pilar",
  "pink",
  "plain",
  "plane",
  "pleasant",
  "plumb",
  "plump",
  "plus",
  "plush",
  "poised",
  "polar",
  "poor",
  "pop",
  "port",
  "posed",
  "posh",
  "potty",
  "powerful",
  "precious",
  "prickly",
  "pricy",
  "prim",
  "prior",
  "privy",
  "prize",
  "prone",
  "proof",
  "prosy",
  "proud",
  "pubic",
  "pudgy",
  "puff",
  "puffy",
  "pulpy",
  "punk",
  "puny",
  "pupal",
  "pure",
  "pushy",
  "putrid",
  "puzzled",
  "quack",
  "quaint",
  "quasi",
  "quick",
  "quiet",
  "rabid",
  "radio",
  "rainy",
  "rapid",
  "rare",
  "rash",
  "raspy",
  "ratty",
  "raw",
  "ready",
  "real",
  "rear",
  "red",
  "regal",
  "relieved",
  "repulsive",
  "retro",
  "rich",
  "rife",
  "right",
  "rigid",
  "riled",
  "ripe",
  "risen",
  "risky",
  "ritzy",
  "roast",
  "robed",
  "rocky",
  "roomy",
  "ropey",
  "rose",
  "rosy",
  "rough",
  "round",
  "rowdy",
  "royal",
  "ruby",
  "rude",
  "ruled",
  "rum",
  "rummy",
  "runic",
  "runny",
  "runty",
  "rural",
  "rush",
  "rushy",
  "rust",
  "rusty",
  "rutty",
  "sad",
  "safe",
  "sage",
  "said",
  "salt",
  "salty",
  "same",
  "sandy",
  "sane",
  "sappy",
  "sassy",
  "saute",
  "saved",
  "scaly",
  "scant",
  "scary",
  "scrub",
  "seamy",
  "sear",
  "seedy",
  "self",
  "selfish",
  "sent",
  "seven",
  "sewed",
  "sewn",
  "shady",
  "shaky",
  "sham",
  "sharp",
  "shed",
  "sheer",
  "shiny",
  "short",
  "shot",
  "showy",
  "shut",
  "shy",
  "sick",
  "side",
  "sign",
  "silky",
  "silly",
  "silty",
  "sissy",
  "six",
  "sixth",
  "sixty",
  "size",
  "sized",
  "skew",
  "skim",
  "slack",
  "slain",
  "slaty",
  "slav",
  "sleek",
  "sleepy",
  "slick",
  "slim",
  "slimy",
  "slow",
  "sly",
  "small",
  "smart",
  "smiling",
  "smoggy",
  "smoky",
  "smug",
  "snaky",
  "sneak",
  "snide",
  "snowy",
  "snub",
  "snuff",
  "snug",
  "soapy",
  "sober",
  "soft",
  "soggy",
  "solar",
  "sold",
  "sole",
  "solid",
  "solo",
  "some",
  "sooty",
  "sore",
  "sorry",
  "sound",
  "soupy",
  "sour",
  "south",
  "sown",
  "spare",
  "sparkling",
  "spent",
  "spicy",
  "spiky",
  "spiny",
  "splay",
  "splendid",
  "split",
  "spotless",
  "spry",
  "spumy",
  "squab",
  "squat",
  "stagy",
  "stale",
  "star",
  "stark",
  "steep",
  "stern",
  "stiff",
  "still",
  "stock",
  "stone",
  "stony",
  "stormy",
  "stout",
  "strange",
  "straw",
  "stray",
  "stuck",
  "stung",
  "stupid",
  "suave",
  "successful",
  "such",
  "sudsy",
  "sulky",
  "sunk",
  "sunny",
  "super",
  "sure",
  "surly",
  "sweet",
  "swell",
  "swept",
  "swift",
  "swish",
  "sworn",
  "tabby",
  "taboo",
  "tacky",
  "taken",
  "talented",
  "talky",
  "tall",
  "tame",
  "tamed",
  "tan",
  "tangy",
  "taped",
  "tardy",
  "tart",
  "tasty",
  "tawny",
  "teal",
  "ten",
  "tender",
  "tenor",
  "tense",
  "tenth",
  "tepid",
  "terrible",
  "terse",
  "testy",
  "thankful",
  "thick",
  "thin",
  "third",
  "thoughtful",
  "thoughtless",
  "three",
  "tidal",
  "tidy",
  "tied",
  "tight",
  "tiled",
  "timed",
  "timid",
  "tinny",
  "tiny",
  "tipsy",
  "tired",
  "toed",
  "token",
  "tonal",
  "toned",
  "tonic",
  "top",
  "tops",
  "torn",
  "total",
  "tough",
  "toxic",
  "tried",
  "trig",
  "trim",
  "trite",
  "troubled",
  "true",
  "tubby",
  "tubed",
  "tumid",
  "twee",
  "twin",
  "two",
  "ugliest",
  "ugly",
  "ultra",
  "uncut",
  "under",
  "undue",
  "unfed",
  "unfit",
  "uninterested",
  "union",
  "unlit",
  "unsightly",
  "unusual",
  "unwed",
  "upper",
  "upset",
  "uptight",
  "urban",
  "used",
  "usual",
  "utter",
  "vague",
  "vain",
  "valid",
  "vapid",
  "vast",
  "victorious",
  "viral",
  "vital",
  "vivacious",
  "vivid",
  "vocal",
  "void",
  "wacky",
  "wandering",
  "warm",
  "wary",
  "washy",
  "waste",
  "wavy",
  "waxed",
  "waxen",
  "waxy",
  "weak",
  "weary",
  "weedy",
  "weeny",
  "weepy",
  "weird",
  "well",
  "welsh",
  "west",
  "wet",
  "whiny",
  "white",
  "whole",
  "wicked",
  "wide",
  "wide-eyed",
  "wild",
  "wily",
  "wimpy",
  "windy",
  "wired",
  "wiry",
  "wise",
  "wispy",
  "witty",
  "wonky",
  "woody",
  "wooly",
  "woozy",
  "wordy",
  "wormy",
  "worn",
  "worried",
  "worrisome",
  "worse",
  "worst",
  "worth",
  "wound",
  "woven",
  "wrong",
  "wroth",
  "wry",
  "young",
  "yucky",
  "yummy",
  "zany",
  "zealous",
  "zero",
  "zesty",
  "zippy",
  "zonal",
]

nouns = [
  "ace",
  "ache",
  "acid",
  "acme",
  "acorn",
  "acre",
  "act",
  "actor",
  "add",
  "adder",
  "adept",
  "advil",
  "afro",
  "agave",
  "age",
  "aged",
  "agent",
  "agony",
  "ailey",
  "aim",
  "aioli",
  "air",
  "aisle",
  "akron",
  "alarm",
  "album",
  "ale",
  "alert",
  "algae",
  "alias",
  "alibi",
  "alien",
  "alley",
  "alloy",
  "ally",
  "aloe",
  "alpha",
  "alps",
  "altar",
  "amber",
  "amigo",
  "amino",
  "amish",
  "ammo",
  "amp",
  "angel",
  "anger",
  "angle",
  "angst",
  "angus",
  "anime",
  "ankle",
  "annex",
  "anole",
  "ant",
  "ante",
  "antic",
  "anvil",
  "ape",
  "apex",
  "aphid",
  "apple",
  "april",
  "apron",
  "aqua",
  "arbor",
  "arc",
  "arch",
  "area",
  "arena",
  "argon",
  "argus",
  "ark",
  "arm",
  "armor",
  "arms",
  "army",
  "aroma",
  "array",
  "arrow",
  "arson",
  "art",
  "ascot",
  "aspen",
  "asset",
  "ate",
  "atom",
  "attic",
  "audio",
  "audit",
  "auger",
  "aunt",
  "aunty",
  "aura",
  "auto",
  "award",
  "awe",
  "awl",
  "axe",
  "axiom",
  "axis",
  "axle",
  "azure",
  "baby",
  "back",
  "bacon",
  "bad",
  "badge",
  "bag",
  "bagel",
  "bail",
  "bait",
  "baker",
  "bale",
  "balk",
  "ball",
  "balm",
  "ban",
  "band",
  "bane",
  "banjo",
  "bank",
  "banks",
  "bar",
  "barb",
  "bard",
  "barge",
  "bark",
  "barn",
  "baron",
  "bars",
  "base",
  "bash",
  "basic",
  "basil",
  "basin",
  "basis",
  "bass",
  "bat",
  "batch",
  "bath",
  "baton",
  "bay",
  "bayou",
  "beach",
  "bead",
  "beads",
  "beak",
  "beam",
  "bean",
  "bear",
  "beard",
  "beast",
  "beat",
  "beats",
  "bed",
  "bee",
  "beech",
  "beef",
  "beep",
  "beer",
  "beet",
  "begin",
  "beige",
  "being",
  "belch",
  "bell",
  "belly",
  "belt",
  "bench",
  "bend",
  "bends",
  "bent",
  "beret",
  "berry",
  "bet",
  "beta",
  "bevel",
  "bevy",
  "bias",
  "bib",
  "bible",
  "bid",
  "bidet",
  "bike",
  "biker",
  "bill",
  "bin",
  "bind",
  "bingo",
  "biome",
  "biped",
  "birch",
  "bird",
  "birth",
  "bison",
  "bit",
  "bite",
  "biter",
  "black",
  "blade",
  "blame",
  "blank",
  "blast",
  "blaze",
  "blend",
  "blimp",
  "blind",
  "bling",
  "blink",
  "blip",
  "bliss",
  "blitz",
  "bloat",
  "blob",
  "block",
  "blog",
  "bloke",
  "blond",
  "blood",
  "bloom",
  "blow",
  "blue",
  "blues",
  "bluff",
  "blur",
  "blurb",
  "blush",
  "boa",
  "boar",
  "board",
  "boast",
  "boat",
  "bod",
  "body",
  "bog",
  "bogey",
  "boil",
  "bold",
  "bolt",
  "bomb",
  "bond",
  "bone",
  "boner",
  "bones",
  "bong",
  "bongo",
  "bonus",
  "boo",
  "book",
  "boom",
  "boon",
  "boost",
  "boot",
  "booth",
  "booty",
  "booze",
  "bore",
  "borer",
  "born",
  "boss",
  "bot",
  "botch",
  "bound",
  "bow",
  "bowel",
  "bowl",
  "bowls",
  "box",
  "boxer",
  "boy",
  "bra",
  "brace",
  "brag",
  "braid",
  "brail",
  "brain",
  "brake",
  "bran",
  "brand",
  "brass",
  "brat",
  "brave",
  "bravo",
  "brawl",
  "brawn",
  "bread",
  "break",
  "breed",
  "brew",
  "briar",
  "bribe",
  "brick",
  "bride",
  "brie",
  "brief",
  "brim",
  "brine",
  "brink",
  "brit",
  "brits",
  "britt",
  "broad",
  "broil",
  "brood",
  "brook",
  "broom",
  "broth",
  "brow",
  "brown",
  "brunt",
  "brush",
  "brute",
  "buck",
  "bud",
  "buddy",
  "budge",
  "buff",
  "bug",
  "buggy",
  "bugle",
  "build",
  "bulb",
  "bulge",
  "bulk",
  "bull",
  "bully",
  "bum",
  "bump",
  "bun",
  "bunch",
  "bung",
  "bunk",
  "bunny",
  "buns",
  "bunt",
  "buoy",
  "bur",
  "burn",
  "burns",
  "burp",
  "burst",
  "bus",
  "bush",
  "buss",
  "bust",
  "buy",
  "buyer",
  "buzz",
  "bye",
  "bylaw",
  "byte",
  "cab",
  "cabin",
  "cable",
  "cabot",
  "cache",
  "caddy",
  "cadet",
  "cafe",
  "cage",
  "cager",
  "cake",
  "calf",
  "call",
  "calm",
  "cam",
  "camel",
  "camp",
  "can",
  "canal",
  "candy",
  "cane",
  "cap",
  "cape",
  "caper",
  "car",
  "carat",
  "card",
  "cards",
  "care",
  "caret",
  "cargo",
  "carp",
  "carry",
  "cart",
  "case",
  "cash",
  "cask",
  "cast",
  "caste",
  "cat",
  "catch",
  "caulk",
  "cause",
  "cave",
  "cavil",
  "caw",
  "cease",
  "cedar",
  "cell",
  "cello",
  "cent",
  "chaff",
  "chain",
  "chair",
  "chalk",
  "champ",
  "chant",
  "chaos",
  "chap",
  "chard",
  "charm",
  "chart",
  "chase",
  "chasm",
  "chat",
  "cheat",
  "check",
  "cheek",
  "cheep",
  "cheer",
  "chef",
  "chess",
  "chest",
  "chew",
  "chic",
  "chick",
  "chief",
  "child",
  "chill",
  "chime",
  "chimp",
  "chin",
  "chip",
  "chips",
  "chirp",
  "chit",
  "chive",
  "chock",
  "choir",
  "choke",
  "choky",
  "chomp",
  "chop",
  "chord",
  "chore",
  "chow",
  "chuck",
  "chug",
  "chum",
  "chump",
  "chunk",
  "churn",
  "chute",
  "cider",
  "cigar",
  "cinch",
  "cite",
  "city",
  "clack",
  "claim",
  "clam",
  "clamp",
  "clams",
  "clan",
  "clang",
  "clank",
  "clap",
  "clash",
  "clasp",
  "class",
  "clay",
  "clean",
  "clear",
  "cleat",
  "cleft",
  "clerk",
  "click",
  "cliff",
  "climb",
  "cling",
  "clip",
  "cloak",
  "clock",
  "clog",
  "clone",
  "close",
  "clot",
  "cloth",
  "cloud",
  "clout",
  "clove",
  "clown",
  "club",
  "cluck",
  "clue",
  "clump",
  "clunk",
  "coach",
  "coal",
  "coast",
  "coat",
  "cobra",
  "cocoa",
  "cod",
  "code",
  "cog",
  "coil",
  "coin",
  "coke",
  "cola",
  "cold",
  "colon",
  "color",
  "colt",
  "coma",
  "comb",
  "combo",
  "come",
  "comet",
  "comic",
  "comma",
  "conch",
  "condo",
  "cone",
  "coney",
  "conk",
  "cook",
  "cool",
  "coot",
  "cop",
  "cope",
  "copy",
  "coral",
  "cord",
  "cords",
  "core",
  "cork",
  "corn",
  "corp",
  "corps",
  "cost",
  "costs",
  "cosy",
  "cot",
  "couch",
  "cough",
  "count",
  "court",
  "cove",
  "coven",
  "cover",
  "cow",
  "cowl",
  "cows",
  "cozy",
  "crab",
  "crabs",
  "crack",
  "craft",
  "cramp",
  "crane",
  "crank",
  "crash",
  "crate",
  "crawl",
  "craze",
  "crazy",
  "creak",
  "cream",
  "cred",
  "cree",
  "creed",
  "creek",
  "creep",
  "crepe",
  "cress",
  "crest",
  "crew",
  "crib",
  "crime",
  "crimp",
  "crisp",
  "croak",
  "crock",
  "crook",
  "crop",
  "cross",
  "crow",
  "crowd",
  "crown",
  "crud",
  "crude",
  "crumb",
  "crush",
  "crust",
  "crux",
  "cry",
  "crypt",
  "cub",
  "cubby",
  "cube",
  "cubit",
  "cue",
  "cuff",
  "cull",
  "cult",
  "cup",
  "curb",
  "curd",
  "cure",
  "curl",
  "curry",
  "curse",
  "curve",
  "cut",
  "cyan",
  "cycle",
  "cynic",
  "dab",
  "daily",
  "dairy",
  "daisy",
  "dame",
  "damp",
  "dance",
  "dandy",
  "dane",
  "dare",
  "dark",
  "dart",
  "darts",
  "dash",
  "data",
  "date",
  "dawn",
  "day",
  "days",
  "daze",
  "dead",
  "deaf",
  "deal",
  "dean",
  "dear",
  "death",
  "debit",
  "debt",
  "debut",
  "decal",
  "decay",
  "deck",
  "decor",
  "decoy",
  "deed",
  "deeds",
  "deep",
  "deer",
  "delay",
  "deli",
  "delta",
  "demo",
  "demon",
  "denim",
  "dent",
  "depot",
  "depth",
  "derby",
  "desk",
  "detox",
  "deuce",
  "devil",
  "dew",
  "dial",
  "diary",
  "dibs",
  "dice",
  "diet",
  "dig",
  "digit",
  "digs",
  "dill",
  "dime",
  "diner",
  "ding",
  "dip",
  "dirt",
  "disc",
  "disco",
  "dish",
  "disk",
  "ditch",
  "ditto",
  "dive",
  "diver",
  "dock",
  "dodge",
  "dog",
  "dogma",
  "doll",
  "dolly",
  "dolt",
  "dome",
  "donor",
  "donut",
  "doom",
  "door",
  "dope",
  "dork",
  "dorm",
  "dot",
  "doubt",
  "dough",
  "dove",
  "dowel",
  "down",
  "dozen",
  "dozer",
  "draft",
  "drag",
  "drain",
  "drama",
  "drape",
  "draw",
  "dread",
  "dream",
  "dress",
  "drew",
  "drier",
  "drift",
  "drill",
  "drink",
  "drip",
  "drive",
  "drone",
  "drool",
  "drop",
  "drove",
  "drug",
  "druid",
  "drum",
  "dry",
  "dryer",
  "duck",
  "duct",
  "due",
  "duel",
  "duet",
  "dug",
  "dunce",
  "dune",
  "dunk",
  "dusk",
  "dust",
  "duty",
  "dye",
  "dyer",
  "dying",
  "eager",
  "eagle",
  "ear",
  "earth",
  "ease",
  "easel",
  "east",
  "eater",
  "eats",
  "echo",
  "edge",
  "eel",
  "egg",
  "eggs",
  "ego",
  "eight",
  "elbow",
  "elder",
  "elect",
  "elf",
  "elite",
  "elk",
  "elm",
  "elves",
  "email",
  "ember",
  "empty",
  "emu",
  "end",
  "enemy",
  "entry",
  "envy",
  "epic",
  "epoxy",
  "equal",
  "era",
  "error",
  "essay",
  "eve",
  "even",
  "event",
  "evil",
  "exam",
  "exile",
  "exit",
  "extra",
  "eye",
  "eyes",
  "fable",
  "face",
  "facet",
  "fact",
  "fad",
  "fade",
  "faint",
  "fair",
  "fairy",
  "faith",
  "fake",
  "fall",
  "falls",
  "fame",
  "fan",
  "fancy",
  "fang",
  "far",
  "farce",
  "fare",
  "farm",
  "fast",
  "fat",
  "fate",
  "fault",
  "favor",
  "fawn",
  "fax",
  "fear",
  "feast",
  "feat",
  "fed",
  "fee",
  "feed",
  "feel",
  "felt",
  "femur",
  "fence",
  "fern",
  "ferry",
  "fetch",
  "feud",
  "fever",
  "few",
  "fib",
  "fiber",
  "field",
  "fiend",
  "fifth",
  "fifty",
  "fig",
  "fight",
  "file",
  "filet",
  "fill",
  "film",
  "filth",
  "final",
  "finch",
  "find",
  "fine",
  "fire",
  "firm",
  "first",
  "fish",
  "fist",
  "fit",
  "five",
  "fiver",
  "fives",
  "fix",
  "fixer",
  "fizz",
  "flag",
  "flair",
  "flak",
  "flake",
  "flame",
  "flank",
  "flap",
  "flaps",
  "flare",
  "flash",
  "flask",
  "flat",
  "flats",
  "flaw",
  "flea",
  "fleet",
  "flesh",
  "flex",
  "flick",
  "flier",
  "flies",
  "fling",
  "flint",
  "flip",
  "flirt",
  "float",
  "flock",
  "flood",
  "floor",
  "flop",
  "floss",
  "flour",
  "flow",
  "flu",
  "flub",
  "fluff",
  "fluid",
  "fluke",
  "flume",
  "flush",
  "flute",
  "flux",
  "fly",
  "flyer",
  "foam",
  "focus",
  "fog",
  "foil",
  "fold",
  "folk",
  "folks",
  "folly",
  "font",
  "food",
  "fool",
  "foot",
  "force",
  "forge",
  "fork",
  "form",
  "fort",
  "forth",
  "forty",
  "forum",
  "foul",
  "found",
  "four",
  "fowl",
  "fox",
  "foyer",
  "frail",
  "frame",
  "frat",
  "fraud",
  "fray",
  "freak",
  "free",
  "freon",
  "fret",
  "friar",
  "fries",
  "frill",
  "frisk",
  "frizz",
  "frog",
  "front",
  "frost",
  "froth",
  "frown",
  "fruit",
  "fry",
  "fryer",
  "fudge",
  "fuel",
  "full",
  "fume",
  "fumes",
  "fun",
  "fund",
  "funds",
  "fungi",
  "funk",
  "funny",
  "fur",
  "fury",
  "fuse",
  "fuss",
  "futon",
  "fuze",
  "fuzz",
  "gag",
  "gage",
  "gain",
  "game",
  "gamma",
  "gap",
  "gape",
  "gas",
  "gash",
  "gasp",
  "gate",
  "gates",
  "gator",
  "gauge",
  "gavel",
  "gawk",
  "gaze",
  "gear",
  "gecko",
  "geek",
  "gel",
  "gem",
  "gene",
  "genie",
  "genoa",
  "genre",
  "gent",
  "germ",
  "ghost",
  "ghoul",
  "giant",
  "gift",
  "gild",
  "gimp",
  "gin",
  "gipsy",
  "girl",
  "gist",
  "give",
  "given",
  "giver",
  "gizmo",
  "glad",
  "glade",
  "gland",
  "glans",
  "glare",
  "glass",
  "glaze",
  "gleam",
  "glee",
  "glide",
  "glint",
  "globe",
  "gloom",
  "glory",
  "gloss",
  "glove",
  "glow",
  "glue",
  "gnat",
  "gnome",
  "goal",
  "goat",
  "going",
  "gold",
  "golem",
  "golf",
  "goner",
  "goo",
  "good",
  "goof",
  "goofy",
  "goon",
  "goose",
  "goth",
  "gouge",
  "gown",
  "grab",
  "grace",
  "grad",
  "grade",
  "graft",
  "grail",
  "grain",
  "gram",
  "grand",
  "grant",
  "grape",
  "graph",
  "grasp",
  "grass",
  "grate",
  "grave",
  "gravy",
  "gray",
  "graze",
  "great",
  "greed",
  "green",
  "grey",
  "grid",
  "grief",
  "grill",
  "grime",
  "grin",
  "grind",
  "grip",
  "gripe",
  "grit",
  "grits",
  "groan",
  "groom",
  "gross",
  "group",
  "grove",
  "growl",
  "grub",
  "gruel",
  "grump",
  "grunt",
  "guard",
  "guess",
  "guest",
  "guide",
  "guild",
  "guilt",
  "gulch",
  "gulf",
  "gull",
  "gulp",
  "gum",
  "gun",
  "guppy",
  "guru",
  "gush",
  "gust",
  "gut",
  "guts",
  "guy",
  "gym",
  "habit",
  "hack",
  "hag",
  "hail",
  "hair",
  "half",
  "hall",
  "halo",
  "halt",
  "ham",
  "hand",
  "hands",
  "handy",
  "hang",
  "hare",
  "harp",
  "hash",
  "haste",
  "hat",
  "hatch",
  "hate",
  "hater",
  "haunt",
  "have",
  "haven",
  "havoc",
  "hawk",
  "hay",
  "haze",
  "hazel",
  "head",
  "heap",
  "heaps",
  "heart",
  "heat",
  "heavy",
  "hedge",
  "heed",
  "heel",
  "heft",
  "heir",
  "helix",
  "hell",
  "hello",
  "helm",
  "help",
  "hem",
  "hemp",
  "hen",
  "herb",
  "herd",
  "here",
  "hero",
  "hex",
  "hick",
  "hide",
  "high",
  "hike",
  "hiker",
  "hill",
  "hilt",
  "hind",
  "hinge",
  "hint",
  "hip",
  "hippo",
  "hippy",
  "hire",
  "hiss",
  "hit",
  "hitch",
  "hive",
  "hives",
  "hoagy",
  "hoard",
  "hoax",
  "hob",
  "hobby",
  "hobo",
  "hog",
  "hoist",
  "hold",
  "hole",
  "home",
  "honey",
  "honk",
  "honor",
  "hoof",
  "hook",
  "hooks",
  "hoop",
  "hoops",
  "hoot",
  "hop",
  "hope",
  "hops",
  "horde",
  "horn",
  "horse",
  "hose",
  "host",
  "hotel",
  "hound",
  "hour",
  "hours",
  "house",
  "howl",
  "hub",
  "hue",
  "huff",
  "hug",
  "hula",
  "hulk",
  "hull",
  "hum",
  "human",
  "humor",
  "hump",
  "humus",
  "hunch",
  "hunk",
  "hunt",
  "hurl",
  "hurry",
  "hurt",
  "hush",
  "husk",
  "husky",
  "hut",
  "hydra",
  "hyena",
  "hymn",
  "hype",
  "ibis",
  "ice",
  "icing",
  "icon",
  "idea",
  "ideal",
  "idiom",
  "idiot",
  "idle",
  "idler",
  "idol",
  "igloo",
  "iglu",
  "ill",
  "image",
  "imp",
  "inch",
  "index",
  "info",
  "ingot",
  "ink",
  "inlet",
  "inn",
  "input",
  "intro",
  "ion",
  "iris",
  "iron",
  "irony",
  "isle",
  "issue",
  "itch",
  "ivory",
  "ivy",
  "jab",
  "jack",
  "jacks",
  "jail",
  "jam",
  "jamb",
  "jar",
  "java",
  "jaw",
  "jay",
  "jazz",
  "jean",
  "jeans",
  "jeep",
  "jeer",
  "jello",
  "jelly",
  "jest",
  "jet",
  "jetty",
  "jewel",
  "jig",
  "jive",
  "job",
  "jock",
  "jog",
  "join",
  "joint",
  "joist",
  "joke",
  "joker",
  "jolly",
  "jolt",
  "joust",
  "joy",
  "judge",
  "jug",
  "juice",
  "juke",
  "jump",
  "junk",
  "junky",
  "juror",
  "jury",
  "kale",
  "kayak",
  "kazoo",
  "kebab",
  "keen",
  "keep",
  "keg",
  "kelp",
  "key",
  "kick",
  "kid",
  "kiddy",
  "kiln",
  "kilo",
  "kilt",
  "kin",
  "kind",
  "king",
  "kiss",
  "kit",
  "kite",
  "kitty",
  "kiwi",
  "knack",
  "knee",
  "kneel",
  "knell",
  "knife",
  "knit",
  "knob",
  "knock",
  "knot",
  "know",
  "koala",
  "krill",
  "lab",
  "label",
  "labor",
  "lace",
  "lack",
  "lad",
  "ladle",
  "lady",
  "lag",
  "lair",
  "lake",
  "lamb",
  "lame",
  "lamp",
  "lance",
  "land",
  "lane",
  "lap",
  "lapel",
  "lapse",
  "lard",
  "large",
  "larva",
  "laser",
  "lash",
  "lass",
  "lasso",
  "last",
  "lat",
  "latch",
  "latex",
  "lathe",
  "latte",
  "laugh",
  "lava",
  "law",
  "lawn",
  "laws",
  "lay",
  "layer",
  "layup",
  "leach",
  "lead",
  "leaf",
  "leak",
  "lean",
  "leap",
  "lear",
  "lease",
  "leash",
  "least",
  "leave",
  "ledge",
  "leech",
  "leeds",
  "leek",
  "leer",
  "left",
  "lefty",
  "leg",
  "lego",
  "legs",
  "lemon",
  "lemur",
  "lens",
  "lent",
  "let",
  "level",
  "lever",
  "liar",
  "libel",
  "lick",
  "lid",
  "lie",
  "lied",
  "life",
  "lift",
  "light",
  "like",
  "lilac",
  "limb",
  "limbo",
  "lime",
  "limit",
  "limp",
  "line",
  "linen",
  "liner",
  "link",
  "links",
  "lint",
  "lion",
  "lip",
  "lisp",
  "list",
  "lit",
  "liter",
  "liver",
  "llama",
  "loach",
  "load",
  "loads",
  "loaf",
  "loan",
  "lob",
  "lobby",
  "lobe",
  "local",
  "lock",
  "lodge",
  "loft",
  "log",
  "logic",
  "logo",
  "loner",
  "look",
  "loom",
  "loon",
  "loony",
  "loop",
  "loot",
  "lord",
  "loser",
  "loss",
  "lost",
  "lot",
  "lots",
  "lotto",
  "lotus",
  "love",
  "lover",
  "low",
  "lower",
  "luck",
  "lump",
  "lunch",
  "lung",
  "lure",
  "lush",
  "lying",
  "mace",
  "macro",
  "madam",
  "mafia",
  "magi",
  "magic",
  "magma",
  "maid",
  "mail",
  "main",
  "major",
  "maker",
  "male",
  "malt",
  "mam",
  "mama",
  "mamba",
  "mambo",
  "mamma",
  "man",
  "mane",
  "mango",
  "mania",
  "manor",
  "map",
  "maple",
  "march",
  "mare",
  "mark",
  "marks",
  "mars",
  "marsh",
  "mash",
  "mask",
  "mass",
  "mast",
  "mat",
  "match",
  "mate",
  "mates",
  "math",
  "maths",
  "max",
  "maxim",
  "may",
  "mayo",
  "mayor",
  "maze",
  "meal",
  "mean",
  "means",
  "meat",
  "medal",
  "medic",
  "meet",
  "meld",
  "melee",
  "melon",
  "melt",
  "memo",
  "men",
  "mend",
  "menu",
  "meow",
  "mercy",
  "merit",
  "mesh",
  "mess",
  "metal",
  "meter",
  "meth",
  "metro",
  "might",
  "mile",
  "milk",
  "mill",
  "mills",
  "mimer",
  "mimic",
  "min",
  "mince",
  "mind",
  "mine",
  "miner",
  "mini",
  "mink",
  "minor",
  "mint",
  "minus",
  "miser",
  "miss",
  "mist",
  "mite",
  "miter",
  "mitt",
  "mix",
  "mixer",
  "moan",
  "moat",
  "mob",
  "mocha",
  "mock",
  "mod",
  "modal",
  "mode",
  "model",
  "modem",
  "mogul",
  "mojo",
  "molar",
  "mold",
  "mole",
  "molt",
  "mom",
  "momma",
  "mommy",
  "money",
  "monk",
  "month",
  "moo",
  "mooch",
  "mood",
  "moody",
  "moon",
  "moose",
  "mop",
  "mope",
  "moped",
  "moral",
  "morse",
  "moss",
  "motel",
  "moth",
  "motor",
  "motto",
  "mould",
  "mound",
  "mount",
  "mouse",
  "mouth",
  "move",
  "mover",
  "movie",
  "mow",
  "mucus",
  "mud",
  "muff",
  "mug",
  "mulch",
  "mule",
  "mum",
  "mummy",
  "munch",
  "mural",
  "muse",
  "mush",
  "music",
  "musk",
  "must",
  "mute",
  "mutt",
  "mylar",
  "nacho",
  "name",
  "namer",
  "names",
  "nanna",
  "nap",
  "nasal",
  "navy",
  "neck",
  "need",
  "needy",
  "neon",
  "nepal",
  "nerd",
  "nerve",
  "nest",
  "net",
  "news",
  "newt",
  "nick",
  "niece",
  "night",
  "nine",
  "niner",
  "ninja",
  "ninth",
  "noble",
  "nod",
  "node",
  "noise",
  "nomad",
  "none",
  "nook",
  "noon",
  "noose",
  "north",
  "nose",
  "notch",
  "note",
  "noun",
  "nudge",
  "nuke",
  "nun",
  "nurse",
  "nut",
  "nylon",
  "oaf",
  "oak",
  "oar",
  "oasis",
  "oat",
  "oates",
  "oath",
  "ocean",
  "octet",
  "odds",
  "ode",
  "odor",
  "offer",
  "ogre",
  "oil",
  "oiler",
  "oink",
  "okay",
  "old",
  "oldie",
  "olive",
  "omega",
  "omen",
  "one",
  "onion",
  "onset",
  "ooze",
  "open",
  "optic",
  "oral",
  "orange",
  "orb",
  "orbit",
  "orca",
  "order",
  "ore",
  "oreo",
  "organ",
  "ounce",
  "out",
  "oval",
  "oven",
  "over",
  "owl",
  "owner",
  "oxbow",
  "oxen",
  "ozone",
  "pace",
  "pacer",
  "pack",
  "pact",
  "pad",
  "page",
  "pager",
  "pail",
  "pain",
  "pains",
  "paint",
  "pair",
  "pal",
  "pale",
  "palm",
  "pan",
  "panda",
  "pane",
  "panel",
  "panic",
  "pansy",
  "pant",
  "pants",
  "papa",
  "paper",
  "par",
  "park",
  "parks",
  "part",
  "parts",
  "party",
  "pass",
  "past",
  "pasta",
  "paste",
  "pat",
  "patch",
  "path",
  "patio",
  "pause",
  "pave",
  "paw",
  "pawn",
  "pay",
  "payer",
  "peace",
  "peach",
  "peak",
  "pear",
  "pearl",
  "pecan",
  "pedal",
  "peek",
  "peel",
  "peer",
  "peg",
  "pelt",
  "pen",
  "penny",
  "perch",
  "peril",
  "perk",
  "pesto",
  "pet",
  "petal",
  "petty",
  "phase",
  "phone",
  "photo",
  "piano",
  "pick",
  "pie",
  "piece",
  "pier",
  "pig",
  "piggy",
  "pigmy",
  "pike",
  "pile",
  "piles",
  "pill",
  "pimp",
  "pin",
  "pinch",
  "pine",
  "ping",
  "pink",
  "pinky",
  "pinot",
  "pint",
  "pipe",
  "pit",
  "pita",
  "pitch",
  "pitt",
  "pity",
  "pivot",
  "pixel",
  "pizza",
  "place",
  "plaid",
  "plain",
  "plan",
  "plane",
  "plank",
  "plant",
  "plate",
  "play",
  "plaza",
  "plea",
  "plier",
  "plot",
  "plow",
  "ploy",
  "pluck",
  "plug",
  "plum",
  "plumb",
  "plume",
  "plump",
  "plus",
  "plush",
  "plyer",
  "pod",
  "poem",
  "poet",
  "point",
  "poke",
  "poker",
  "pole",
  "poll",
  "polls",
  "pond",
  "pong",
  "pony",
  "pooch",
  "poof",
  "pool",
  "poor",
  "pop",
  "poppy",
  "porch",
  "pore",
  "pork",
  "port",
  "pose",
  "poser",
  "post",
  "pot",
  "pouch",
  "pound",
  "power",
  "prank",
  "prawn",
  "press",
  "prey",
  "price",
  "pride",
  "prime",
  "prism",
  "prize",
  "pro",
  "probe",
  "prom",
  "promo",
  "proof",
  "prop",
  "props",
  "prose",
  "prowl",
  "prune",
  "pry",
  "pub",
  "puck",
  "puff",
  "pug",
  "pull",
  "pulp",
  "pulse",
  "puma",
  "pump",
  "pun",
  "punch",
  "punk",
  "punks",
  "punt",
  "pup",
  "pupil",
  "puppy",
  "purge",
  "purse",
  "push",
  "put",
  "putt",
  "putty",
  "quack",
  "quad",
  "quake",
  "qualm",
  "quart",
  "queen",
  "query",
  "quest",
  "quick",
  "quid",
  "quiet",
  "quilt",
  "quirk",
  "quirt",
  "quiz",
  "quota",
  "quote",
  "race",
  "racer",
  "rad",
  "radar",
  "radio",
  "raft",
  "rafts",
  "rag",
  "rage",
  "raid",
  "rail",
  "rails",
  "rain",
  "raise",
  "rake",
  "rally",
  "ram",
  "ramp",
  "ranch",
  "range",
  "rank",
  "rant",
  "rap",
  "rapid",
  "rash",
  "rat",
  "rate",
  "rates",
  "ratio",
  "raw",
  "ray",
  "razor",
  "razz",
  "reach",
  "read",
  "ready",
  "real",
  "realm",
  "ream",
  "rear",
  "rebel",
  "red",
  "reed",
  "reef",
  "reek",
  "reel",
  "reign",
  "relay",
  "relic",
  "rent",
  "reply",
  "reset",
  "resin",
  "rest",
  "retro",
  "revel",
  "rhino",
  "rhyme",
  "rib",
  "rice",
  "ricer",
  "rich",
  "ride",
  "rider",
  "ridge",
  "riff",
  "rifle",
  "rift",
  "rig",
  "right",
  "rim",
  "rind",
  "ring",
  "rings",
  "rink",
  "rinse",
  "riot",
  "rip",
  "rise",
  "riser",
  "risk",
  "rite",
  "rival",
  "river",
  "roach",
  "road",
  "roads",
  "roar",
  "roast",
  "robe",
  "robin",
  "robot",
  "rock",
  "rod",
  "rodeo",
  "rogue",
  "role",
  "roll",
  "room",
  "rooms",
  "roost",
  "root",
  "roots",
  "rope",
  "rose",
  "rot",
  "rotor",
  "rouge",
  "rough",
  "round",
  "route",
  "rover",
  "row",
  "rowdy",
  "rower",
  "royal",
  "rub",
  "rube",
  "ruby",
  "rug",
  "rugby",
  "ruin",
  "rule",
  "ruler",
  "rum",
  "rummy",
  "rumor",
  "run",
  "rune",
  "rung",
  "runt",
  "ruse",
  "rush",
  "rust",
  "rut",
  "saber",
  "safe",
  "sag",
  "saga",
  "sage",
  "sail",
  "saint",
  "salad",
  "sale",
  "salem",
  "sales",
  "salon",
  "salsa",
  "salt",
  "same",
  "sand",
  "sands",
  "sang",
  "sash",
  "sass",
  "sauce",
  "sauna",
  "save",
  "saver",
  "savor",
  "saw",
  "say",
  "scale",
  "scan",
  "scar",
  "scare",
  "scarf",
  "scene",
  "scent",
  "scold",
  "scone",
  "scoop",
  "scope",
  "score",
  "scorn",
  "scout",
  "scrap",
  "sea",
  "seal",
  "seam",
  "seat",
  "seats",
  "sect",
  "sedan",
  "see",
  "seed",
  "seek",
  "seer",
  "self",
  "sell",
  "sense",
  "serum",
  "serve",
  "servo",
  "set",
  "setup",
  "seven",
  "shack",
  "shade",
  "shake",
  "sham",
  "shame",
  "shank",
  "shape",
  "shard",
  "share",
  "shark",
  "sharp",
  "shave",
  "shawl",
  "shed",
  "sheep",
  "sheet",
  "shelf",
  "shell",
  "shift",
  "shill",
  "shim",
  "shin",
  "ship",
  "shirt",
  "shoe",
  "shoes",
  "shop",
  "shore",
  "shot",
  "shove",
  "show",
  "shred",
  "shrub",
  "shrug",
  "shy",
  "sick",
  "siege",
  "sigh",
  "sight",
  "sign",
  "silk",
  "silks",
  "silly",
  "silo",
  "sin",
  "sink",
  "sinus",
  "sip",
  "sir",
  "siren",
  "six",
  "sixer",
  "sixth",
  "sixty",
  "size",
  "ski",
  "skid",
  "skier",
  "skill",
  "skim",
  "skin",
  "skip",
  "skirt",
  "skit",
  "skull",
  "skunk",
  "sky",
  "slab",
  "slack",
  "slag",
  "slain",
  "slam",
  "slang",
  "slant",
  "slap",
  "slash",
  "slate",
  "slave",
  "slaw",
  "sled",
  "sleep",
  "sleet",
  "slew",
  "slews",
  "slice",
  "slick",
  "slide",
  "slime",
  "sling",
  "slip",
  "slit",
  "slob",
  "slope",
  "slot",
  "sloth",
  "slug",
  "slum",
  "slump",
  "slur",
  "slush",
  "smack",
  "small",
  "smart",
  "smash",
  "smear",
  "smell",
  "smelt",
  "smile",
  "smirk",
  "smith",
  "smock",
  "smog",
  "smoke",
  "snack",
  "snag",
  "snail",
  "snake",
  "snap",
  "snare",
  "snarl",
  "sneak",
  "sniff",
  "snipe",
  "snore",
  "snort",
  "snot",
  "snow",
  "snug",
  "soak",
  "soap",
  "soar",
  "sob",
  "sock",
  "sofa",
  "softy",
  "soil",
  "sole",
  "solid",
  "son",
  "sonar",
  "song",
  "sonny",
  "soot",
  "sooth",
  "sore",
  "sort",
  "soul",
  "sound",
  "soup",
  "sour",
  "south",
  "spa",
  "space",
  "spade",
  "spam",
  "span",
  "spar",
  "spare",
  "spark",
  "spasm",
  "spat",
  "spawn",
  "speed",
  "spell",
  "spelt",
  "spice",
  "spike",
  "spill",
  "spin",
  "spit",
  "spite",
  "splat",
  "split",
  "spoil",
  "spoke",
  "spoof",
  "spook",
  "spool",
  "spoon",
  "spore",
  "sport",
  "spot",
  "spots",
  "spout",
  "spray",
  "spree",
  "spud",
  "spur",
  "spurt",
  "spy",
  "squat",
  "squid",
  "stab",
  "stack",
  "staff",
  "stag",
  "stage",
  "stain",
  "stair",
  "stake",
  "stalk",
  "stall",
  "stamp",
  "stand",
  "star",
  "stare",
  "start",
  "stash",
  "state",
  "stay",
  "stays",
  "steak",
  "steal",
  "steam",
  "steed",
  "steel",
  "steer",
  "stem",
  "step",
  "steps",
  "stern",
  "stew",
  "stick",
  "stiff",
  "still",
  "stilt",
  "sting",
  "stink",
  "stint",
  "stir",
  "stock",
  "stoic",
  "stomp",
  "stone",
  "stool",
  "stoop",
  "stop",
  "stops",
  "store",
  "stork",
  "storm",
  "story",
  "stove",
  "strap",
  "straw",
  "stray",
  "strip",
  "strum",
  "strut",
  "stub",
  "stud",
  "study",
  "stuff",
  "stump",
  "stunt",
  "style",
  "sub",
  "suds",
  "sugar",
  "suit",
  "suite",
  "sum",
  "sumer",
  "sun",
  "sung",
  "super",
  "surf",
  "surge",
  "sushi",
  "sutra",
  "swab",
  "swag",
  "swamp",
  "swan",
  "swap",
  "swarm",
  "sway",
  "sweat",
  "sweep",
  "sweet",
  "swell",
  "swift",
  "swim",
  "swine",
  "swing",
  "swipe",
  "swirl",
  "swish",
  "syrup",
  "table",
  "tack",
  "taco",
  "tact",
  "tad",
  "taffy",
  "tag",
  "tail",
  "tails",
  "take",
  "taker",
  "tale",
  "talk",
  "talks",
  "tall",
  "tally",
  "talon",
  "tan",
  "tank",
  "tap",
  "tape",
  "taps",
  "tar",
  "tarp",
  "tart",
  "task",
  "taste",
  "taunt",
  "tax",
  "taxer",
  "taxi",
  "taxis",
  "tea",
  "teach",
  "teal",
  "team",
  "tear",
  "tears",
  "tease",
  "teen",
  "teens",
  "teeth",
  "tell",
  "temp",
  "tempo",
  "ten",
  "tense",
  "tent",
  "tenth",
  "term",
  "terms",
  "test",
  "text",
  "thaw",
  "theft",
  "theme",
  "then",
  "there",
  "theta",
  "thick",
  "thief",
  "thigh",
  "thing",
  "think",
  "third",
  "thorn",
  "three",
  "throw",
  "thud",
  "thug",
  "thumb",
  "tick",
  "tide",
  "tidy",
  "tie",
  "tier",
  "tiger",
  "tilde",
  "tile",
  "till",
  "time",
  "timer",
  "times",
  "timid",
  "tin",
  "tint",
  "tip",
  "tire",
  "titan",
  "title",
  "toad",
  "toady",
  "toast",
  "today",
  "toe",
  "toil",
  "token",
  "toll",
  "tomb",
  "tome",
  "ton",
  "tone",
  "toner",
  "tongs",
  "tonic",
  "tons",
  "tool",
  "toon",
  "toot",
  "tooth",
  "top",
  "topic",
  "torch",
  "torso",
  "toss",
  "total",
  "tote",
  "totem",
  "touch",
  "tough",
  "tour",
  "tours",
  "tow",
  "towel",
  "tower",
  "town",
  "towny",
  "toxin",
  "toy",
  "trace",
  "track",
  "trade",
  "trail",
  "train",
  "trait",
  "trap",
  "trash",
  "tray",
  "tread",
  "treat",
  "tree",
  "trek",
  "trend",
  "triad",
  "trial",
  "trick",
  "trim",
  "trio",
  "trip",
  "troll",
  "troop",
  "trot",
  "trout",
  "truce",
  "truck",
  "true",
  "trump",
  "trunk",
  "trust",
  "truth",
  "try",
  "tub",
  "tuba",
  "tube",
  "tuck",
  "tug",
  "tulip",
  "tummy",
  "tumor",
  "tuna",
  "tune",
  "tuner",
  "tunic",
  "turf",
  "turn",
  "tush",
  "tusk",
  "tutor",
  "twine",
  "twins",
  "twirl",
  "twist",
  "two",
  "tying",
  "type",
  "typo",
  "udder",
  "ulcer",
  "uncle",
  "union",
  "unit",
  "unity",
  "upper",
  "upset",
  "urn",
  "usage",
  "use",
  "user",
  "usher",
  "using",
  "valet",
  "valor",
  "value",
  "valve",
  "van",
  "vase",
  "vat",
  "vault",
  "vegan",
  "veil",
  "vein",
  "venom",
  "vent",
  "venue",
  "verb",
  "verge",
  "vest",
  "vet",
  "vial",
  "vibe",
  "vibes",
  "vice",
  "video",
  "view",
  "vigil",
  "vine",
  "vinyl",
  "viola",
  "viper",
  "virgo",
  "virus",
  "visit",
  "visor",
  "vista",
  "vocal",
  "vodka",
  "vogue",
  "voice",
  "void",
  "volt",
  "vote",
  "voter",
  "vow",
  "vowel",
  "wacko",
  "wad",
  "wade",
  "wader",
  "wads",
  "wafer",
  "waft",
  "wag",
  "wage",
  "wager",
  "wages",
  "wagon",
  "wail",
  "wain",
  "waist",
  "wait",
  "wake",
  "walk",
  "wall",
  "waltz",
  "wane",
  "want",
  "war",
  "ward",
  "ware",
  "warp",
  "wart",
  "wash",
  "wasp",
  "waste",
  "watch",
  "water",
  "watt",
  "watts",
  "wave",
  "waver",
  "wax",
  "way",
  "ways",
  "wear",
  "weave",
  "web",
  "wed",
  "wedge",
  "week",
  "weird",
  "well",
  "wells",
  "welsh",
  "west",
  "wet",
  "whack",
  "whale",
  "wharf",
  "wheat",
  "wheel",
  "whey",
  "whiff",
  "while",
  "whim",
  "whip",
  "whirl",
  "whisk",
  "white",
  "who",
  "whole",
  "whore",
  "why",
  "wick",
  "widow",
  "width",
  "wife",
  "wig",
  "wild",
  "will",
  "wilt",
  "wimp",
  "win",
  "wince",
  "winch",
  "wind",
  "wine",
  "wing",
  "wings",
  "wink",
  "wipe",
  "wiper",
  "wire",
  "wise",
  "wish",
  "wit",
  "witch",
  "wits",
  "woe",
  "wolf",
  "woman",
  "womb",
  "won",
  "wood",
  "woods",
  "woof",
  "wool",
  "word",
  "words",
  "work",
  "works",
  "world",
  "worm",
  "worry",
  "worse",
  "worst",
  "wort",
  "worth",
  "wound",
  "wow",
  "wrack",
  "wrap",
  "wrath",
  "wreck",
  "wring",
  "wrist",
  "wrong",
  "yam",
  "yard",
  "yarn",
  "yawn",
  "yay",
  "year",
  "years",
  "yeast",
  "yell",
  "yes",
  "yeti",
  "yield",
  "yoga",
  "yolk",
  "young",
  "youth",
  "zap",
  "zebra",
  "zinc",
  "zing",
  "zip",
  "zit",
  "zone",
  "zoo",
  "zoom",
  "zero",
  "zany",
  "whir",
  "welt",
  "whig",
  "wand",
  "twin",
  "tribe",
  "tilt",
  "sword",
  "spine",
  "spear",
  "site",
  "shock",
  "sent",
]

def memorable_id():
  return f"{random.choice(adjectives)}-{random.choice(adjectives)}-{random.choice(nouns)}"




from pathlib import Path
from ..data.io import load_json
from ..data import flatten
from .supervised import Trainer
from ..callbacks import Callback


def step(model, inputs, targets, optimizer, loss, callback: Callback = None):
  model.train()
  optimizer.zero_grad()

  pred = model(inputs)
  loss = loss(pred, targets)

  loss.backward()
  optimizer.step()

  return inputs, targets, pred, loss


def train(model, batches, optimizer, loss, device=None, step=step, callback: Callback=None):
  for inputs, targets in batches:
    if device:
      inputs, targets = inputs.to(device), targets.to(device)
    yield step(model, inputs, targets, optimizer, loss)


def collect_summaries(root='.', name='summary.json', pandas=True):
  s = [load_json(f) for f in Path(root).glob(f'**/*{name}')]
  if pandas:
    import pandas as pd
    return pd.DataFrame([flatten(x) for x in s])
  else:
    return s


import pyarrow


ctx = pyarrow.serialization.default_serialization_context()

def enable_torch_serialization(context=ctx):
  pyarrow.serialization.register_torch_serialization_handlers(
    context
  )

# enable by default
enable_torch_serialization()


def deserialize_arrow(buf, context=ctx):
  return pyarrow.deserialize(buf, context)


def serialize_arrow(x, context=ctx):
  return pyarrow.serialize(x, context=context).to_buffer()


# use arrow as default serialization method
serialize = serialize_arrow
deserialize = deserialize_arrow


def to_bytes(string: str) -> bytes:
  return string.encode(encoding='utf-8', errors='strict')


def to_unicode(b: bytes) -> str:
  return b.decode(encoding='utf-8', errors='strict')


from PIL import Image
import io


def image_to_bytes(image: Image.Image, format='jpeg'):
  buff = io.BytesIO()
  image.save(buff, format=format)
  return buff.getvalue()


def image_from_bytes(buffer):
  return Image.open(io.BytesIO(buffer))


def enable_loading_truncated_images():
  from PIL import ImageFile
  ImageFile.LOAD_TRUNCATED_IMAGES = True



from mypy_extensions import TypedDict
from typing import List, Optional

"""
/process/
/info/
/feedback/
/inputs/id/

POST /predict/

{
  inputs: [
    {
      id: 'asfsdfasf',
      base64: 'data:/asfsda
    }
  ]
  params: {
    word_threshold: .5,
    rotate: true,
    width: 400,
    vocab: [] 
  }
  
  meta: {
    client_id: 124123,
    session_id: 3414,
    
  }
}



{
  request_id: 'afdfa',
  results: [
    {
      id: 'asfsda',
      index: 0,
      text_lines: []
      boxes: [],
      transcript: []
    }
  ],
  params: {
    size: 800,
    correct: true
  },
  meta: {
    start_time: 141242,
    end_time: 31312,
  }
}
"""

UUID = str

class Meta(TypedDict):
  pass

class Input(TypedDict):
  id: Optional[UUID]
  meta: Optional[Meta]

class Base64Image(Input):
  base64: str


class Result(TypedDict):
  id: UUID
  index: int

class Error(TypedDict):
  id: UUID


class Request(TypedDict):
  request_id: Optional[UUID]
  inputs: List[Input]
  params: dict
  meta: Meta


class Response(TypedDict):
  model_id: UUID
  request_id: UUID

  results: Optional[List[Result]]
  errors: Optional[List[Error]]

  params: dict

  meta: Meta

class Feedback:
  type: str


def endpoint(f):
  def foo():
    return f()
  return f

class Server:
  input = []
  output = []

  def __init__(self, model, preprocess):
    self.model = model

  async def preprocess(self):
    pass

  async def process(self):
    pass

  async def postprocess(self):
    pass

  def __call__(self):
    pass

  async def socket(self):
    pass

class ClassificationServer(Server):
  pass

class OCRServer(Server):
  @endpoint
  async def detect(self):
    pass

import yann


yann.serve = lambda **kwargs: kwargs
yann.serve(
  checkpoint='gs://model/asfadf',
  model='resnet50',
  preprocess='imagenet',
  classes='imagenet',
  port=8080,
  workers=4,
  device=yann.default.device
)


import json
import os
import pickle as pkl
import tarfile
from collections import namedtuple
import csv
import gzip
from pathlib import Path
import torch


class Loader:
  """
    gs://bucket/file.th
    ./foo/**/*.jpg
  """

  def __call__(self, path, format=None, deserialize=None, filesystem=None, **kwargs):
    path = Path(path)
    format = format or path.suffix[1:]
    if hasattr(self, format):
      return getattr(self, format)(str(path), **kwargs)
    raise ValueError(f'File format not supported ({format})')

  def th(self, path, **kwargs):
    return torch.load(path, **kwargs)

  def json(self, path, **kwargs):
    return load_json(path, **kwargs)

  def pickle(self, path, **kwargs):
    return load_pickle(path, **kwargs)

  def parquet(self, path, **kwargs):
    import pandas as pd
    return pd.read_parquet(path, **kwargs)

  def csv(self, path, **kwargs):
    import pandas as pd
    return pd.read_csv(path, **kwargs)

  def tsv(self, path, **kwargs):
    import pandas as pd
    return pd.read_csv(path, **kwargs)

  pkl = pickle


load = Loader()


class Saver:
  def __call__(
    self, x, path, format=None, serialize=None, filesystem=None, **kwargs
  ):
    path = Path(path)
    format = format or path.suffix[1:]
    if hasattr(self, format):
      return getattr(self, format)(x, path, **kwargs)
    raise ValueError(f'File format not supported ({format})')

  def th(self, x, path, **kwargs):
    return torch.save(x, path, **kwargs)

  def json(self, x, path, **kwargs):
    return save_json(x, path, **kwargs)

  def pickle(self, x, path, **kwargs):
    return save_pickle(x, path, **kwargs)

  pkl = pickle


save = Saver()


def save_pickle(obj, path, mode='wb'):
  with open(str(path), mode) as f:
    pkl.dump(obj, f)


def load_pickle(path, mode='rb'):
  with open(str(path), mode) as f:
    return pkl.load(f)


def save_json(obj, path, mode='w'):
  with open(str(path), mode) as f:
    json.dump(obj, f)
  return path


def load_json(path, mode='r'):
  with open(str(path), mode) as f:
    return json.load(f)


def tar_dir(path, dest=None, zip=True):
  path = str(path)
  dest = str(dest or path)

  ext, mode = ('.tar.gz', 'w:gz') if zip else ('.tar', 'w')

  if not dest.endswith(ext):
    dest = os.path.splitext(dest)[0] + ext

  with tarfile.open(dest, mode) as tar:
    tar.add(path)


def lines(path, mode='r'):
  with open(path, mode=mode) as f:
    yield from f


def write_lines(items, path, mode='w'):
  with open(path, mode=mode) as f:
    f.writelines(items)


def untar(path):
  with tarfile.open(path) as tar:
    tar.extractall()


def unzip(zip, dest):
  import zipfile
  with zipfile.ZipFile(zip, 'r') as f:
      f.extractall(dest)


def iter_csv(path, header=True, tuples=True, sep=',', quote='"', **kwargs):
  with open(path) as f:
    reader = csv.reader(f, delimiter=sep, quotechar=quote, **kwargs)
    if header:
      if tuples:
        reader = iter(reader)
        h = next(reader)
        Row = namedtuple('Row', h)

        for r in reader:
          yield Row(*r)
      else:
        reader = iter(reader)
        h = next(reader)
        for r in reader:
          yield dict(zip(h, r))
    else:
      yield from reader


def write_csv(data, path, header=None):
  with gzip.open(path, 'wt') if path.endswith('.gz') else open(path, 'w') as f:
    writer = csv.writer(f)
    if header:
      writer.writerow(header)
    for row in data:
      writer.writerow(row)



import datetime
import logging
import pathlib
import torch
import torch.nn
import types
from statistics import mean
from torch.optim.optimizer import Optimizer
from torch.utils.data import Sampler
from typing import Optional, Callable

from yann.utils import fully_qualified_name
from yann.data.loaders import DataLoader
from .. import callbacks as yann_callbacks
from .. import resolve, evaluate, to, default, trainable
from ..callbacks import get_callbacks
from ..data import get_dataset_name, Classes
from ..data.io import save_json
from ..datasets import TransformDataset
from ..export import export
from ..inference.predict import Classifier
from ..params import HyperParams
from ..train.base import BaseTrainer
from ..utils import counter, print_tree, timestr, hash_params
from ..utils.decorators import lazy
from ..utils.ids import memorable_id


def get_model_name(model):
  if isinstance(model, torch.nn.DataParallel):
    model = model.module

  if hasattr(model, 'name'):
    return model.name

  return model.__class__.__name__


class Paths:
  # TODO: add a way to display the directory tree
  def __init__(self, root):
    self.root = pathlib.Path(root)

    self.checkpoint_format = '{}{}'

  def create(self):
    self.root.mkdir(parents=True, exist_ok=True)

  @property
  def checkpoints(self):
    path = self.root / 'checkpoints'
    path.mkdir(parents=True, exist_ok=True)
    return path

  def checkpoint(self, **kwargs):
    return self.checkpoints / self.checkpoint_format.format(**kwargs)

  @property
  def tensorboard(self):
    return self.root / 'tensorboard'

  @property
  def logs(self):
    return self.root / 'logs'

  @property
  def evals(self):
    return self.root / 'evals'

  @property
  def plots(self):
    return self.root / 'plots'

  @property
  def outputs(self):
    return self.root / 'outputs'

  @property
  def exports(self):
    return self.root / 'exports'

  @property
  def summary(self):
    return self.root / 'summary.json'

  def tree(self, **kwargs):
    print_tree(self.root, **kwargs)


class Trainer(BaseTrainer):
  model: torch.nn.Module
  loss: Callable
  optimizer: Optional[Optimizer]

  lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler]
  loader: DataLoader
  classes: Optional[Classes]
  sampler: Optional[Sampler]
  params: Optional[HyperParams]
  paths: Paths



  def __init__(
      self,
      model=None,
      dataset=None,
      optimizer=None,
      loss=None,
      loader=None,
      sampler=None,
      num_workers=8,
      transform=None,
      transform_batch=None,
      lr_scheduler=None,
      lr_batch_step=False,
      callbacks=None,
      device=None,
      parameters='trainable',
      batch_size=16,
      val_dataset=None,
      val_loader=None,
      val_transform=None,
      classes=None,
      parallel=False,
      name=None,
      description=None,
      root='./experiments/train-runs/',
      metrics=None,
      collate=None,
      params=None,
      pin_memory=True,
      step=None,
      id=None
  ):
    """

    Args:
      model: model to train
      dataset:
      optimizer:
      loss:
      loader:
      sampler:
      num_workers:
      transform:
      transform_batch: transform function that augments a batch of data
      lr_scheduler:
      lr_batch_step: if true will call optimizer.step() after each batch,
        otherwise will call it at the end of each epoch
      callbacks:
      device:
      parameters:
      batch_size:
      val_dataset:
      val_loader:
      val_transform:
      classes:
      parallel:
      name:
      description:
      root:
      metrics:
      collate:
      params:
      pin_memory:
      step:
      id:
    """
    super().__init__()

    self.id = id or memorable_id()

    self.params = params

    self.model = resolve.model(
      model,
      required=True,
      validate=callable
    )
    if parallel:
      self.model = torch.nn.DataParallel(self.model)

    self.loss = resolve.loss(
      loss,
      required=True,
      validate=callable,
    )

    if parameters == 'trainable' and self.model:
      parameters = trainable(self.model.parameters())
    self.optimizer = resolve.optimizer(
      optimizer,
      args=(parameters or self.model.parameters(),),
      required=True,
      validate=lambda x: hasattr(x, 'step')
    )

    self.dataset = resolve.dataset(
      dataset,
      required=not loader,
    )

    if classes:
      self.classes = (
        classes if isinstance(classes, Classes)
        else Classes(classes)
      )
    elif hasattr(self.dataset, 'classes') and isinstance(self.dataset.classes, Classes):
      self.classes = self.dataset.classes
    else:
      self.classes = None

    if transform:
      self.dataset = TransformDataset(self.dataset, transform)
    self.transform = transform

    self.transform_batch = transform_batch

    self.batch_size = batch_size

    self.loader = loader or DataLoader(
      self.dataset,
      batch_size=self.batch_size,
      pin_memory=pin_memory,
      shuffle=False if sampler else True,
      sampler=sampler,
      num_workers=num_workers,
      **({'collate_fn': collate} if collate else {})
    )

    self.val_dataset = resolve.dataset(
      val_dataset,
      required=False,
    )

    self.val_transform = val_transform or transform
    if self.val_transform:
      self.val_dataset = TransformDataset(self.val_dataset, self.val_transform)

    self.val_loader = val_loader or (val_dataset and DataLoader(
      self.val_dataset,
      batch_size=batch_size,
      shuffle=False,
      pin_memory=True,
      num_workers=num_workers
    ))

    self.lr_scheduler = resolve.lr_scheduler(
      lr_scheduler,
      kwargs=dict(optimizer=self.optimizer)
    )
    self.lr_batch_step = lr_batch_step

    if step is not None:
      self.override(self.step, step)

    self.num_samples = 0
    self.num_steps = 0
    self.num_epochs = 0

    self.time_created = datetime.datetime.utcnow()

    self.name = name or (
      f"{get_dataset_name(self.loader)}-{get_model_name(self.model)}"
    )
    self.description = description

    self.paths = Paths(pathlib.Path(root)  / self.name / timestr(self.time_created))
    self.paths.create()

    self.history = None
    has_history = False
    if callbacks:
      if callbacks is True:
        # if callbacks=True use the default set of callbacks
        callbacks = get_callbacks()
      for c in callbacks:
        if isinstance(c, yann_callbacks.History):
          self.history = c
          has_history = True
          break
    metrics = (metrics,) if isinstance(metrics, str) else metrics
    self.history = self.history or yann_callbacks.History(*(metrics or ()))

    self.callbacks = callbacks or [
      yann_callbacks.Logger()
    ]

    if not has_history:
      self.callbacks.append(self.history)

    # make sure history callback is called first
    self.callbacks = sorted(
      self.callbacks,
      key=lambda x: 0 if x is self.history else 1)

    self.function_callback = None

    device = default.device if device is None else device
    if device:
      self.device = torch.device(device) if isinstance(device, str) else device
      self.to(self.device)

  @property
  def root(self):
    """for backwards compatibility, self.paths.root used to be on self.root"""
    return self.paths.root

  def __setattr__(self, key, value):
    if key == 'optimizer':
      if hasattr(self, 'lr_scheduler') and hasattr(self.lr_scheduler, 'optimizer'):
        self.lr_scheduler.optimizer = value
    if key == 'loader':
      if hasattr(self, 'dataset') and hasattr(value, 'dataset'):
        super(Trainer, self).__setattr__('dataset', value.dataset)
      if hasattr(value, 'batch_size'):
        super(Trainer, self).__setattr__('batch_size', value.batch_size)
    if key == 'batch_size' and hasattr(self, 'batch_size') and self.batch_size != key:
      if hasattr(self, 'loader') and self.loader:
        raise ValueError(
          'Cannot modify batch_size because a loader is defined '
          'and modifying batch size of a loader is not supported, '
          'try creating and setting a new loader instead'
        )
      if key == 'dataset' and hasattr(self, 'dataset') and self.dataset \
          and hasattr(self, 'loader') and self.loader:
        raise ValueError(
          'Cannot modify dataset because a loader is defined '
          'and modifying dataset of a loader is not supported, '
          'try creating and setting a new loader instead'
        )

    logging.debug(f"setting '{key}' to {value}")
    super(Trainer, self).__setattr__(key, value)

  # @classmethod
  # def from_params(cls, params):
  #   pass

  def to(self, device=None):
    self.device = device

    to(
      self.model,
      self.loss,
      self.optimizer,
      device=self.device
    )

  def on(self, event, callback=None):
    if not self.function_callback:
      self.function_callback = yann_callbacks.FunctionCallback()
      self.callbacks.append(self.function_callback)

    if callback:
      self.function_callback.on(event, callback)
      return self
    else:
      def decorated(func):
        self.function_callback.on(event, func)
        return func

      return decorated

  def train_mode(self):
    self.model.train()

  def eval_mode(self):
    self.model.eval()

  def epochs(self, num=None):
    """
    Yields current epoch count and keeps internal epoch count
    """
    for e in counter(start=self.num_epochs, end=self.num_epochs + num):
      yield e
      self.num_epochs += 1

  def batches(self, device=None):
    device = device or self.device

    for batch in self.loader:
      if self.transform_batch:
        batch = self.transform_batch(*batch)

      if device:
        yield to(*batch, device=device)
      else:
        yield batch

      self.num_steps += 1
      self.num_samples += len(batch[0])

  def override(self, method, function=False):
    """
    Override a method of the trainer
    Args:
      method: str or method reference
      function: function to be used as a replacement for the given method
    """
    method = method if isinstance(method, str) else method.__name__
    if not hasattr(self, method):
      raise AttributeError(f"Can't override method '{method}' because it's not defined")
    if function is False:
      # assume it's used as a decorator
      # @train.override('step')
      # def custom_step(trainer, inputs, targets):
      def decorator(f):
        setattr(self, method, types.MethodType(f, self))
      return decorator
    else:
      setattr(self, method, types.MethodType(function, self))

  def step(self, inputs, target):
    """

    Args:
      inputs:
      target:

    Returns:

    """
    if not self.model.training:
      self.model.train()

    self.optimizer.zero_grad()

    outputs = self.model(inputs)
    loss = self.loss(outputs, target)
    loss.backward()
    self.optimizer.step()

    return outputs, loss

  def validate(self, loader=None, device=None):
    loader = loader or self.val_loader
    device = device or self.device

    self.model.eval()

    self.on_validation_start()

    ts, os = [], []
    with torch.no_grad():
      for inputs, targets, outputs in evaluate(self.model, loader, device):
        self.on_validation_batch(inputs=inputs, targets=targets,
                                 outputs=outputs)
        ts.append(targets)
        os.append(outputs)

      ts = torch.cat(ts)
      os = torch.cat(os)
      loss = self.loss(os, ts)
    self.on_validation_end(targets=ts, outputs=os, loss=loss)
    return loss

  def run(self, epochs=None):
    self._stop = False
    try:
      self.on_train_start()

      for _ in self.epochs(num=epochs):
        self.on_epoch_start(epoch=self.num_epochs)
        for batch_idx, (inputs, targets) in enumerate(self.batches()):
          self.on_batch_start(
            batch=batch_idx,
            inputs=inputs,
            targets=targets
          )
          try:
            outputs, loss = self.step(inputs, targets)
          except KeyboardInterrupt as e:
            self.stop()
            break
          except Exception as e:
            self.on_batch_error(batch=batch_idx, error=e)
            if self._stop: break
            raise e

          self.on_batch_end(
            batch=batch_idx,
            inputs=inputs,
            targets=targets,
            outputs=outputs,
            loss=loss
          )

          if self.lr_scheduler and self.lr_batch_step:
            self.lr_scheduler.step(epoch=self.num_steps)

          if self._stop: break
        if self._stop: break

        if self.val_loader:
          val_loss = self.validate()
          if self.lr_scheduler and not self.lr_batch_step:
            if 'metrics' in self.lr_scheduler.step.__code__.co_varnames:
             self.lr_scheduler.step(metrics=val_loss, epoch=self.num_epochs)
            else:
              self.lr_scheduler.step(epoch=self.num_epochs)
        elif self.lr_scheduler and not self.lr_batch_step:
          if 'metrics' in self.lr_scheduler.step.__code__.co_varnames:
            loss = mean(self.history.metrics['loss'][-20:])
            self.lr_scheduler.step(metrics=loss, epoch=self.num_epochs)
          else:
            self.lr_scheduler.step(epoch=self.num_epochs)

        self.on_epoch_end(epoch=self.num_epochs)

      self.on_train_end()
    except Exception as e:
      self.on_error(e)
      raise e

  def __call__(self, *args, **kwargs):
    self.run(*args, **kwargs)

  def stop(self):
    self._stop = True

  @lazy
  def predict(self):
    return Classifier(
      model=self.model,
      classes=self.classes,
      preprocess=self.transform,
      postprocess=None
    )

  def checkpoint(self, name=None) -> pathlib.Path:
    state = self.state_dict()
    path = self.paths.checkpoints / (
      f"{name}.th" if name else
      f"{timestr()}-epoch-{self.num_epochs:03d}-steps-{self.num_steps:05d}.th"
    )
    torch.save(state, str(path))
    return path

  def load_checkpoint(self, path, metadata=True):
    # TODO: add 'latest', 'best' support
    data = torch.load(path)
    self.load_state_dict(data, metadata=metadata)

  def export(self, path=None, trace=False, meta=None, postprocess=None):
    path = path or self.paths.exports / timestr()
    export(
      model=self.model,
      preprocess=self.val_transform,
      postprocess=postprocess,
      classes=self.classes,
      trace=trace and next(iter(self.loader)),
      path=path,
      meta=dict(
        name=self.name,
        root=str(self.paths.root),
        dataset=get_dataset_name(self.dataset),
        num_steps=self.num_steps,
        **(meta or {})
      )
    )

    return path

  def state_dict(self):
    data = {
      'metadata': {
        'id': self.id,
        'num_steps': self.num_steps,
        'num_samples': self.num_samples,
        'num_epochs': self.num_epochs,
        'batch_size': self.batch_size,
        'name': self.name,
        'time_created': self.time_created,
        'param_hash': hash_params(self.model)
      }
    }

    for k, v in self.__dict__.items():
      if hasattr(v, 'state_dict'):
        data[k] = {
          'state_dict': v.state_dict(),
          'class': fully_qualified_name(v)
        }

    return data

  def load_state_dict(self, data, metadata=True):
    """
    TODO: add a way to specify which parts should be loaded (ex: model only)
    """
    skipped = set()
    for k, v in data.items():
      if 'state_dict' in v and hasattr(self, k):
        getattr(self, k).load_state_dict(v['state_dict'])
        logging.debug(f"loaded {k}")
      else:
        skipped.add(k)

    if metadata and 'metadata' in data:
      skipped.discard('metadata')
      for k, v in data['metadata'].items():
        setattr(self, k, v)

    if skipped:
      logging.warning(f'skipped {skipped} when loading checkpoint')

  def summary(self):
    summary = {
      'id': self.id,
      'name': self.name,
      'root': str(self.root),
      'num_steps': self.num_steps,
      'num_samples': self.num_samples,
      'num_epochs': self.num_epochs,
      'time_created': str(self.time_created),
      'params': dict(self.params) if self.params else {},
      'metrics': {
        'train': {
          k: ({'min': min(v), 'max': max(v)} if len(v) else {}) for k, v in self.history.metrics.items()
        },
        'validation': {
          k: ({'min': min(v), 'max': max(v)} if len(v) else {}) for k, v in self.history.val_metrics.items()
        }
      }
    }

    summary['duration'] = None
    if len(self.history.metrics.times) > 1:
      try:
        summary['duration'] = self.history.metrics.times[-1] - self.history.metrics.times[0]
      except:
        pass

    return summary

  def save_summary(self):
    save_json(self.summary(), self.paths.summary)
    return self.paths.summary

  def __str__(self):
    return f"""
id: {self.id}
name: {self.name}
root: {self.root}
batch_size: {self.batch_size}
device: {self.device}

MODEL
=====

{self.model}


DATASET
=======

{self.loader.dataset}


LOADER
======

{self.loader}

LOSS
====

{self.loss}


OPTIMIZER
=========

{self.optimizer}

SCHEDULER
=========

{self.lr_scheduler}


PROGRESS
========
epochs: {self.num_epochs}
steps: {self.num_steps}
samples: {self.num_samples}
"""

  def __repr__(self):
    return (
      f"Trainer("
      f"\n  id={self.id},"
      f"\n  name={self.name},"
      f"\n  root={self.root},"
      f"\n  batch_size={self.batch_size},"
      f"\n  device={self.device}"
      "\n)"
    )


"""



sched = Linear(0, 10)

sched[1]


s = Sequence(
  Linear(0, 10, steps=100),

)


for size in sched:
  print(size)

"""


class Schedule:
  def __init__(self, steps=None):
    self.current_step = -1
    self.steps = steps

  def __len__(self):
    return self.steps

  def __getitem__(self, step):
    raise NotImplementedError()

  def __call__(self):
    self.current_step += 1


class Sequence(Schedule):
  def __init__(self, *schedules):
    self.schedules = list(schedules)

  def __len__(self):
    return sum((len(s) for s in self.schedules))

  def __getitem__(self, step):
    start = 0
    for s in self.schedules:
      if step < start + len(s):
        return s[step - start]

      start += len(s)

    raise IndexError()


class Constant(Schedule):
  def __init__(self, value, steps=None):
    self.value = value
    self.steps = steps

  def __getitem__(self, step):
    if step > self.steps:
      raise IndexError()
    return self.value


class Linear(Schedule):
  def __init__(self, start, end, steps):
    self.start = start
    self.end = end
    self.steps = steps

  def __getitem__(self, step):
    if step > self.steps:
      raise IndexError()
    return (self.end - self.start) / (self.steps - 1) * step


# class Lambda(Schedule):
#   def __init__(self, func):
#     pass
#
# class Random(Schedule):
#   pass
#
# class Warmup(Schedule):
#   pass
#
# class Restart(Schedule):
#   pass
#
#
# class Stop(Schedule):
#   """
#
#     sched = Stop(
#       sched,
#       lambda step: train.history.loss[-1] < 0
#     )
#   """
#   pass